# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import pandas as pd
import numpy as np
from tqdm import tqdm
import itertools
import concurrent.futures

from ..functions.coord import load_coord, get_seq_lst, sup_without_map, calc_rmsd
from ..functions.table import fix_val, order_rows, lst_col, merge_dicts
from ..functions.lst import res_to_lst, lst_unique
from ..functions.path import save_matrix, save_json
from ..functions.col import core_path_col, modelid_col, chainid_col


def add_index_dict(final_dict, index_dict):

    for coord_path in list(index_dict.keys()):
        if coord_path not in list(final_dict.keys()):
            final_dict[coord_path] = dict()
        for modelid in list(index_dict[coord_path].keys()):
            if modelid not in list(final_dict[coord_path].keys()):
                final_dict[coord_path][modelid] = dict()
            for chainid in list(index_dict[coord_path][modelid].keys()):
                final_dict[coord_path][modelid][chainid] = index_dict[coord_path][
                    modelid
                ][chainid]

    return final_dict


def build_coord_dict(coord_path):

    return {coord_path: load_coord(coord_path)}


def build_seq_dict(structure, coord_path, modelid, chainid):

    modelid = fix_val(modelid, return_int=True)

    return {coord_path: {modelid: {chainid: get_seq_lst(structure[modelid][chainid])}}}


def build_sup_dict(
    coord_path,
    modelid,
    chainid,
    ref_chain,
    mob_chain,
    ref_seq_lst=None,
    mob_seq_lst=None,
    sup_resid_lst=None,
    pair_aln=True,
):

    sup = sup_without_map(
        ref_chain,
        mob_chain,
        ref_seq_lst=ref_seq_lst,
        mob_seq_lst=mob_seq_lst,
        sup_resids=sup_resid_lst,
        pair_aln=pair_aln,
    )

    return {coord_path: {str(modelid): {chainid: sup[1]}}}


def check_rmsd_dict(
    rmsd_dict, coord_path_1, modelid_1, chainid_1, coord_path_2, modelid_2, chainid_2
):

    rmsd_dist = None
    if coord_path_1 in list(rmsd_dict.keys()):
        if str(modelid_1) in list(rmsd_dict[coord_path_1].keys()):
            if chainid_1 in list(rmsd_dict[coord_path_1][str(modelid_1)].keys()):
                if coord_path_2 in list(
                    rmsd_dict[coord_path_1][str(modelid_1)][chainid_1].keys()
                ):
                    if modelid_2 in list(
                        rmsd_dict[coord_path_1][str(modelid_1)][chainid_1][
                            coord_path_2
                        ].keys()
                    ):
                        if chainid_2 in list(
                            rmsd_dict[coord_path_1][str(modelid_1)][chainid_1][
                                coord_path_2
                            ][str(modelid_2)].keys()
                        ):
                            rmsd_dist = rmsd_dict[coord_path_1][str(modelid_1)][
                                chainid_1
                            ][coord_path_2][str(modelid_2)][chainid_2]

    return rmsd_dist


def append_rmsd_dict(
    rmsd_dict,
    rmsd_dist,
    coord_path_1,
    modelid_1,
    chainid_1,
    coord_path_2,
    modelid_2,
    chainid_2,
):

    if coord_path_1 not in list(rmsd_dict.keys()):
        rmsd_dict[coord_path_1] = dict()
    if str(modelid_1) not in list(rmsd_dict[coord_path_1].keys()):
        rmsd_dict[coord_path_1][str(modelid_1)] = dict()
    if chainid_1 not in list(rmsd_dict[coord_path_1][str(modelid_1)].keys()):
        rmsd_dict[coord_path_1][str(modelid_1)][chainid_1] = dict()
    if coord_path_2 not in list(
        rmsd_dict[coord_path_1][str(modelid_1)][chainid_1].keys()
    ):
        rmsd_dict[coord_path_1][str(modelid_1)][chainid_1][coord_path_2] = dict()
    if str(modelid_2) not in list(
        rmsd_dict[coord_path_1][str(modelid_1)][chainid_1][coord_path_2].keys()
    ):
        rmsd_dict[coord_path_1][str(modelid_1)][chainid_1][coord_path_2][
            str(modelid_2)
        ] = dict()

    rmsd_dict[coord_path_1][str(modelid_1)][chainid_1][coord_path_2][str(modelid_2)][
        chainid_2
    ] = rmsd_dist

    return rmsd_dict


def prep_rmsd_job(
    i_df,
    j_df,
    i_index,
    j_index,
    coord_dict,
    seq_dict=None,
    sup_resid_lst=None,
    rmsd_resid_lst=None,
    rmsd_dict=None,
    coord_path_col=None,
):

    if coord_path_col is None:
        coord_path_col = core_path_col

    i_coord_path = i_df.at[i_index, coord_path_col]
    i_modelid = i_df.at[i_index, modelid_col]
    i_chainid = i_df.at[i_index, chainid_col]

    j_coord_path = j_df.at[j_index, coord_path_col]
    j_modelid = j_df.at[j_index, modelid_col]
    j_chainid = j_df.at[j_index, chainid_col]

    i_chain = coord_dict[i_coord_path][fix_val(i_modelid, return_int=True)][i_chainid]
    j_chain = coord_dict[j_coord_path][fix_val(j_modelid, return_int=True)][j_chainid]

    if seq_dict is not None:
        i_seq_lst = seq_dict[i_coord_path][fix_val(i_modelid, return_int=True)][
            i_chainid
        ]
        j_seq_lst = seq_dict[j_coord_path][fix_val(j_modelid, return_int=True)][
            j_chainid
        ]
    else:
        i_seq_lst = None
        j_seq_lst = None

    rmsd_dist = None
    if rmsd_dict is not None:
        rmsd_dist = check_rmsd_dict(
            rmsd_dict,
            i_coord_path,
            i_modelid,
            i_chainid,
            j_coord_path,
            j_modelid,
            j_chainid,
        )

        if rmsd_dist is None:
            rmsd_dist = check_rmsd_dict(
                rmsd_dict,
                j_coord_path,
                j_modelid,
                j_chainid,
                i_coord_path,
                i_modelid,
                i_chainid,
            )

    return (
        i_chain,
        j_chain,
        i_seq_lst,
        j_seq_lst,
        sup_resid_lst,
        rmsd_resid_lst,
        rmsd_dist,
    )


def calc_rmsd_dist(
    i,
    j,
    i_chain,
    j_chain,
    i_seq_lst=None,
    j_seq_lst=None,
    sup_resids=None,
    rmsd_resids=None,
    rmsd_atomids="CA",
    pair_aln=True,
    rmsd_dist=None,
):

    if rmsd_dist is None:
        sup = sup_without_map(
            i_chain,
            j_chain,
            ref_seq_lst=i_seq_lst,
            mob_seq_lst=j_seq_lst,
            sup_resids=sup_resids,
            pair_aln=pair_aln,
        )

        rmsd_dist = calc_rmsd(
            sup[0],
            sup[1],
            map_dict=sup[2],
            rmsd_resids=rmsd_resids,
            rmsd_atomids=rmsd_atomids,
        )

    return (i, j, rmsd_dist)


def build_rmsd_matrix(
    fit_df,
    rmsd_matrix_path,
    pred_df=None,
    sup_resids=None,
    rmsd_resids=None,
    rmsd_atomids="CA",
    pair_aln=True,
    rmsd_dict=None,
    rmsd_json_path=None,
    coord_path_col=None,
    num_cpu=1,
):

    if coord_path_col is None:
        coord_path_col = core_path_col

    sup_resid_lst = None
    if sup_resids is not None:
        sup_resid_lst = res_to_lst(sup_resids)

    rmsd_resid_lst = None
    if rmsd_resids is not None:
        rmsd_resid_lst = res_to_lst(rmsd_resids)

    fit_df = order_rows(fit_df)

    j_df = fit_df.copy(deep=True)

    coord_path_lst = lst_col(fit_df, coord_path_col, unique=True)

    coord_df = fit_df.copy(deep=True)

    if pred_df is None:
        i_df = fit_df.copy(deep=True)
    else:
        pred_df = order_rows(pred_df)
        i_df = pred_df.copy(deep=True)

        coord_path_lst += lst_col(pred_df, coord_path_col, unique=True)
        coord_df = pd.concat([coord_df, pred_df])

    i_index_lst = list(i_df.index.values)
    j_index_lst = list(j_df.index.values)

    if pred_df is not None:
        index_pairs = itertools.product(i_index_lst, j_index_lst)
    else:
        index_pairs = itertools.combinations(i_index_lst, 2)

    coord_df = coord_df.reset_index(drop=True)

    coord_path_lst = lst_unique(coord_path_lst)

    coord_dict = dict()

    if num_cpu == 1:

        for coord_path in tqdm(
            coord_path_lst, desc="Loading coordinates", position=0, leave=True
        ):
            coord_dict = merge_dicts([coord_dict, build_coord_dict(coord_path)])

    else:

        with concurrent.futures.ProcessPoolExecutor(max_workers=num_cpu) as executor:

            job_lst = [
                executor.submit(build_coord_dict, coord_path)
                for coord_path in coord_path_lst
            ]

            for job in tqdm(
                concurrent.futures.as_completed(job_lst),
                desc="Loading coordinates",
                total=len(job_lst),
                miniters=1,
                position=0,
                leave=True,
            ):

                coord_dict = merge_dicts([coord_dict, job.result()])

    seq_dict = None

    if pair_aln:

        seq_dict = dict()

        if num_cpu == 1:

            for index in tqdm(
                list(coord_df.index.values),
                desc="Loading sequences",
                position=0,
                leave=True,
            ):

                coord_path = coord_df.at[index, coord_path_col]
                modelid = coord_df.at[index, modelid_col]
                chainid = coord_df.at[index, chainid_col]

                structure = coord_dict[coord_path]

                seq_dict = add_index_dict(
                    seq_dict,
                    build_seq_dict(
                        structure,
                        coord_path,
                        modelid,
                        chainid,
                    ),
                )

        else:

            with concurrent.futures.ProcessPoolExecutor(
                max_workers=num_cpu
            ) as executor:

                job_lst = list()

                for index in list(coord_df.index.values):

                    coord_path = coord_df.at[index, coord_path_col]
                    modelid = coord_df.at[index, modelid_col]
                    chainid = coord_df.at[index, chainid_col]

                    structure = coord_dict[coord_path]

                    job_lst.append(
                        executor.submit(
                            build_seq_dict,
                            structure,
                            coord_path,
                            modelid,
                            chainid,
                        )
                    )

                for job in tqdm(
                    concurrent.futures.as_completed(job_lst),
                    desc="Loading sequences",
                    total=len(job_lst),
                    miniters=1,
                    position=0,
                    leave=True,
                ):

                    seq_dict = add_index_dict(seq_dict, job.result())

    matrix = np.zeros((len(i_index_lst), len(j_index_lst)))

    if num_cpu == 1:

        for i_index, j_index in tqdm(
            index_pairs, desc="Building RMSD matrix", position=0, leave=True
        ):

            (
                i_chain,
                j_chain,
                i_seq_lst,
                j_seq_lst,
                sup_resid_lst,
                rmsd_resid_lst,
                rmsd_dist,
            ) = prep_rmsd_job(
                i_df,
                j_df,
                i_index,
                j_index,
                coord_dict,
                seq_dict=seq_dict,
                sup_resid_lst=sup_resid_lst,
                rmsd_resid_lst=rmsd_resid_lst,
                rmsd_dict=rmsd_dict,
                coord_path_col=coord_path_col,
            )

            result = calc_rmsd_dist(
                i_index,
                j_index,
                i_chain,
                j_chain,
                i_seq_lst=i_seq_lst,
                j_seq_lst=j_seq_lst,
                sup_resids=sup_resid_lst,
                rmsd_resids=rmsd_resid_lst,
                rmsd_atomids=rmsd_atomids,
                pair_aln=pair_aln,
                rmsd_dist=rmsd_dist,
            )

            i_index = result[0]
            j_index = result[1]
            dist = result[2]

            matrix[i_index, j_index] = dist

            if pred_df is None:
                matrix[j_index, i_index] = dist

    else:

        with concurrent.futures.ProcessPoolExecutor(max_workers=num_cpu) as executor:

            job_lst = list()
            for i_index, j_index in index_pairs:

                (
                    i_chain,
                    j_chain,
                    i_seq_lst,
                    j_seq_lst,
                    sup_resid_lst,
                    rmsd_resid_lst,
                    rmsd_dist,
                ) = prep_rmsd_job(
                    i_df,
                    j_df,
                    i_index,
                    j_index,
                    coord_dict,
                    seq_dict=seq_dict,
                    sup_resid_lst=sup_resid_lst,
                    rmsd_resid_lst=rmsd_resid_lst,
                    rmsd_dict=rmsd_dict,
                    coord_path_col=coord_path_col,
                )

                job_lst.append(
                    executor.submit(
                        calc_rmsd_dist,
                        i_index,
                        j_index,
                        i_chain,
                        j_chain,
                        i_seq_lst=i_seq_lst,
                        j_seq_lst=j_seq_lst,
                        sup_resids=sup_resid_lst,
                        rmsd_resids=rmsd_resid_lst,
                        rmsd_atomids=rmsd_atomids,
                        pair_aln=pair_aln,
                        rmsd_dist=rmsd_dist,
                    )
                )

            for job in tqdm(
                concurrent.futures.as_completed(job_lst),
                desc="Building RMSD matrix",
                total=len(job_lst),
                miniters=1,
                position=0,
                leave=True,
            ):

                result = job.result()

                i_index = result[0]
                j_index = result[1]
                dist = result[2]

                matrix[i_index, j_index] = dist

                if pred_df is None:
                    matrix[j_index, i_index] = dist

    save_matrix(rmsd_matrix_path, matrix)

    if rmsd_json_path is not None:

        if rmsd_dict is None:
            rmsd_dict = dict()

        if pred_df is not None:
            index_pairs = itertools.product(i_index_lst, j_index_lst)
        else:
            index_pairs = itertools.combinations(i_index_lst, 2)

        if num_cpu == 1:

            for i_index, j_index in tqdm(
                index_pairs, desc="Building RMSD dictionary", position=0, leave=True
            ):
                i_coord_path = i_df.at[i_index, coord_path_col]
                i_modelid = i_df.at[i_index, modelid_col]
                i_chainid = i_df.at[i_index, chainid_col]

                j_coord_path = j_df.at[j_index, coord_path_col]
                j_modelid = j_df.at[j_index, modelid_col]
                j_chainid = j_df.at[j_index, chainid_col]

                rmsd_dist = matrix[i_index, j_index]

                rmsd_dict = append_rmsd_dict(
                    rmsd_dict,
                    rmsd_dist,
                    i_coord_path,
                    i_modelid,
                    i_chainid,
                    j_coord_path,
                    j_modelid,
                    j_chainid,
                )

        else:

            with concurrent.futures.ProcessPoolExecutor(
                max_workers=num_cpu
            ) as executor:

                if pred_df is not None:
                    index_pairs = itertools.product(i_index_lst, j_index_lst)
                else:
                    index_pairs = itertools.combinations(i_index_lst, 2)

                job_lst = list()

                for i_index, j_index in index_pairs:
                    i_coord_path = i_df.at[i_index, coord_path_col]
                    i_modelid = i_df.at[i_index, modelid_col]
                    i_chainid = i_df.at[i_index, chainid_col]

                    j_coord_path = j_df.at[j_index, coord_path_col]
                    j_modelid = j_df.at[j_index, modelid_col]
                    j_chainid = j_df.at[j_index, chainid_col]

                    rmsd_dist = matrix[i_index, j_index]

                    job_lst.append(
                        executor.submit(
                            append_rmsd_dict,
                            rmsd_dict,
                            rmsd_dist,
                            i_coord_path,
                            i_modelid,
                            i_chainid,
                            j_coord_path,
                            j_modelid,
                            j_chainid,
                        )
                    )

                for job in tqdm(
                    concurrent.futures.as_completed(job_lst),
                    desc="Building RMSD dictionary",
                    total=len(job_lst),
                    miniters=1,
                    position=0,
                    leave=True,
                ):

                    rmsd_dict = job.result()

        save_json(rmsd_json_path, rmsd_dict)

    print("Built RMSD matrix!")



# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

from tqdm import tqdm

from ..functions.seq import load_record_dict, get_record_desc
from ..functions.lst import lst_to_str, res_to_lst, str_to_lst, lst_unique, sort_lst
from ..functions.coord import (
    load_coord,
    get_neighbors,
    get_resnum,
    get_chain_cont,
    get_reschainid,
    is_aa,
)
from ..functions.col import (
    core_path_col,
    modelid_col,
    chainid_col,
    pdb_code_col,
    bound_prot_chainid_col,
    bound_prot_col,
    bound_prot_site_col,
    bound_prot_pfam_col,
    bound_prot_swiss_id_col,
    bound_prot_cont_col
)
from ..functions.table import fix_val
from ..functions.path import save_table
from ..functions.pdbaa import get_pdbaa_prot, get_pdbaa_swiss_id


def annot_prot(
    df,
    pdbaa_fasta_path,
    pfam_dict=None,
    site_dict=None,
    other_pfam="Other",
    other_site="Unclassified",
    none_pfam="None",
    none_site="None",
    max_site_dist=4,
    prot_table_path=None,
):

    pdbaa_dict = load_record_dict(pdbaa_fasta_path)

    if site_dict is not None:
        for site_name, site_cont in site_dict.items():
            site_dict[site_name] = res_to_lst(site_cont)

    if pfam_dict is not None:
        pfam_dict_lst = list(pfam_dict.keys())

    for index in tqdm(
        list(df.index.values), desc="Annotating proteins", position=0, leave=True
    ):

        pdb_code = df.at[index, pdb_code_col]

        bound_prot_chainid = df.at[index, bound_prot_chainid_col]

        prot_lst = list()
        cont_lst = list()
        swiss_id_lst = list()
        site_lst = list()
        pfam_lst = list()

        if bound_prot_chainid != "None":
            coord_path = df.at[index, core_path_col]
            modelid = df.at[index, modelid_col]
            chainid = df.at[index, chainid_col]
            structure = load_coord(coord_path)
            neighbors = get_neighbors(
                structure[fix_val(modelid, return_int=True)][chainid]
            )
            for bound_chainid in str_to_lst(bound_prot_chainid):
                record = pdbaa_dict[f"{pdb_code.upper()}{bound_chainid}"]
                desc = get_record_desc(record)
                prot = get_pdbaa_prot(desc)
                swiss_id = get_pdbaa_swiss_id(desc)

                prot_lst.append(prot)
                swiss_id_lst.append(swiss_id)

                resid_cont_lst = lst_unique(
                    [
                        get_resnum(x)
                        for x in get_chain_cont(
                            neighbors,
                            structure[fix_val(modelid, return_int=True)][
                                bound_chainid
                            ],
                            level="R",
                            max_dist=max_site_dist,
                        )
                        if (get_resnum(x) < 50000)
                        and (get_reschainid(x) == chainid)
                        and is_aa(x)
                    ]
                )

                if len(resid_cont_lst) > 0:
                    resid_cont_str = bound_chainid
                    resid_cont_str += ':'
                    resid_cont_str += lst_to_str(resid_cont_lst)
                    cont_lst.append(resid_cont_str)

                if site_dict is not None:

                    max_cont = 0
                    prot_site = other_site
                    for site_name, site_cont_lst in site_dict.items():
                        site_cont = len([x for x in site_cont_lst if x in resid_cont_lst])
                        if site_cont > max_cont:
                            prot_site = site_name
                            max_cont = site_cont
                    site_lst.append(prot_site)

                if pfam_dict is not None:
                    pfam_status = other_pfam
                    if swiss_id in pfam_dict_lst:
                        pfam_status = pfam_dict[swiss_id]
                    elif swiss_id == "NA":
                        pfam_status = "Binder"
                    pfam_lst.append(pfam_status)

        df.at[index, bound_prot_col] = lst_to_str(sort_lst(lst_unique(prot_lst)))
        df.at[index, bound_prot_swiss_id_col] = lst_to_str(
            sort_lst(lst_unique(swiss_id_lst))
        )
        df.at[index, bound_prot_site_col] = lst_to_str(
            sort_lst(lst_unique(site_lst)), join_txt="|", empty=none_site
        )
        df.at[index, bound_prot_pfam_col] = lst_to_str(
            sort_lst(lst_unique(pfam_lst)), join_txt="|", empty=none_pfam
        )

        df.at[index, bound_prot_cont_col] = lst_to_str(cont_lst, join_txt="|")

    print("Annotated proteins!")

    if prot_table_path is not None:
        save_table(prot_table_path, df)
    else:
        return df


# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import numpy as np


def format_lst(val_lst, return_str=False, return_int=False, return_float=False):

    return_type = type(val_lst)

    return return_type(
        [
            format_val(
                x,
                return_str=return_str,
                return_int=return_int,
                return_float=return_float,
            )
            for x in val_lst
        ]
    )


def format_val(val, return_str=False, return_int=False, return_float=False):

    if type(val) == list or type(val) == tuple:
        val = format_lst(
            val, return_str=return_str, return_int=return_int, return_float=return_float
        )
    else:
        if return_str and not return_int and not return_float:
            val = str(val)
        elif return_int and not return_str and not return_float:
            try:
                val = int(val)
            except Exception:
                pass
        elif return_float and not return_str and not return_int:
            try:
                val = float(val)
            except Exception:
                pass

    return val


def format_nested_lst(val_lst, return_str=False, return_int=False, return_float=False):

    for i, val in enumerate(val_lst):
        val_lst[i] = [
            format_val(
                x,
                return_str=return_str,
                return_int=return_int,
                return_float=return_float,
            )
            for x in val
        ]

    return val_lst


def type_lst(data, return_str=False, return_int=False, return_float=False, sort=False):

    if type(data) != list:
        data = list([data])

    data = format_lst(
        data, return_str=return_str, return_int=return_int, return_float=return_float
    )

    if sort:
        data = sorted(data)

    return data


def lst_unique(starting_lst, return_str=False, return_int=False, return_float=False):

    val_lst = list()
    [val_lst.append(x) for x in starting_lst if x not in val_lst]

    val_lst = format_lst(
        val_lst, return_str=return_str, return_int=return_int, return_float=return_float
    )

    return val_lst


def move_end_lst(val_lst, end_lst):

    end_lst = type_lst(end_lst)

    for end in end_lst:
        if end in val_lst:
            val_lst.append(val_lst.pop(val_lst.index(end)))

    return val_lst


def sort_lst(val_lst, return_str=False, return_int=False, return_float=False):

    val_dict = dict()
    for index, val in enumerate(val_lst):
        try:
            fix_val = int(val)
        except:
            fix_val = str(val)

        val_lst[index] = fix_val
        val_dict[fix_val] = val

    val_lst = sorted(
        val_lst,
        key=lambda v: (
            isinstance(v, str),
            v,
        ),
    )

    val_lst = [val_dict[x] for x in val_lst]

    val_lst = format_lst(
        val_lst, return_str=return_str, return_int=return_int, return_float=return_float
    )

    return val_lst


def add_lsts(lst_1, lst_2, return_str=False, return_int=False, return_float=False):

    val_lst = list(set(lst_1 + lst_2))

    val_lst = format_lst(
        val_lst, return_str=return_str, return_int=return_int, return_float=return_float
    )

    return val_lst


def subtract_lsts(lst_1, lst_2, return_str=False, return_int=False, return_float=False):

    val_lst = list(set(lst_1) - set(lst_2))

    val_lst = format_lst(
        val_lst, return_str=return_str, return_int=return_int, return_float=return_float
    )

    return val_lst


def lst_inter(lst_1, lst_2, return_str=False, return_int=False, return_float=False):

    val_lst = list(set(lst_1).intersection((set(lst_2))))

    val_lst = format_lst(
        val_lst, return_str=return_str, return_int=return_int, return_float=return_float
    )

    return val_lst


def lst_diff(
    right_lst, left_lst, return_str=False, return_int=False, return_float=False
):

    val_lst = list(set(left_lst).difference((set(right_lst))))

    val_lst = format_lst(
        val_lst, return_str=return_str, return_int=return_int, return_float=return_float
    )

    return val_lst


def calc_jaccard(lst_1, lst_2, return_dist=False):

    inter = len(lst_inter(lst_1, lst_2))
    union = (len(lst_1) + len(lst_2)) - inter

    jaccard = float(inter) / union

    if return_dist:
        jaccard = 1 - jaccard

    return jaccard


def calc_simpson(lst_1, lst_2, return_dist=False):

    intersect = len([x for x in lst_1 if x in lst_2])

    min_size = np.min(np.array([len(lst_1), len(lst_2)]))

    simpson = intersect / min_size

    if return_dist:
        simpson = 1 - simpson

    return simpson


def lst_nums(start, end, return_str=False, return_int=False, return_float=False):

    first = int(start)
    last = int(end) + 1

    num_lst = list(range(first, last))

    num_lst = format_lst(
        num_lst,
        return_str=return_str,
        return_int=return_int,
        return_float=return_float,
    )

    return num_lst


def lst_to_str(val_lst, join_txt=",", empty=None):

    val_lst = type_lst(val_lst)

    if len(val_lst) == 0:
        val_str = str(empty)
    elif len(val_lst) == format_val(val_lst[0], return_str=True) == "None":
        val_str = str(empty)
    else:
        if len(val_lst) == 0:
            val_str = str(empty)
        elif len(val_lst) > 0:

            val_lst = format_lst(val_lst, return_str=True)
            val_str = str(join_txt.join(val_lst))

    return val_str


def str_to_lst(
    val_str,
    sep_txt=",",
    return_str=False,
    return_int=False,
    return_float=False,
):

    val_str = format_val(val_str, return_str=True)

    if sep_txt in val_str:
        val_lst = list(val_str.split(sep_txt))
    else:
        val_lst = type_lst(val_str)

    val_lst = format_lst(
        val_lst,
        return_str=return_str,
        return_int=return_int,
        return_float=return_float,
    )

    return val_lst


def res_to_str(res_lst, sep_txt=":"):

    if res_lst is None:
        res_str = "all"
    else:
        res_lst = type_lst(res_lst, return_int=True, sort=True)
        res_str = ""
        count = 0

        prev_res = None

        for index, curr_res in enumerate(res_lst):

            if prev_res is not None:
                if index == (len(res_lst) - 1):
                    res_str += f"-{curr_res}"
                else:
                    if (curr_res - prev_res) == 1:
                        count += 1
                    else:
                        if count != 0:
                            res_str += f"-{prev_res}"
                        res_str += sep_txt
                        count = 0

            if count == 0 and index != (len(res_lst) - 1):
                res_str += str(curr_res)

            prev_res = curr_res

    return res_str


def res_to_lst(res_str, return_str=False):

    if res_str == None or type(res_str) == list:
        res_lst = res_str
    elif type(res_str) == int:
        res_lst = type_lst(res_str)
    elif type(res_str) == str:
        if ":" in res_str:
            val_lst = res_str.split(":")
        else:
            val_lst = type_lst(res_str)

        res_lst = list()

        for val in val_lst:
            if "-" in val:
                res_range = val.split("-")
                res_start = res_range[0]
                res_end = res_range[1]
                res_lst += lst_nums(res_start, res_end)
            else:
                res_lst.append(int(val))

    return_int = True
    if return_str:
        return_int = False

    if res_lst is not None:
        res_lst = format_lst(res_lst, return_str=return_str, return_int=return_int)

    return res_lst


def build_range_lst(range, step, sep="-", type=float, dec=2):

    range_lst = list()
    if sep in str(range):
        sep_range = range.split(sep)
        start = type(sep_range[0])
        end = type(sep_range[1])
        while start <= end:
            range_lst.append(start)
            start += type(step)
        if end not in range_lst:
            range_lst.append(end)
    else:
        range_lst.append(type(range))

    if type == float:
        range_lst = [round(x, dec) for x in range_lst]

    return lst_unique(range_lst)


def get_lst_val_indices(lst, val):

    i = -1
    loc_lst = []
    while True:
        try:
            loc = lst.index(val, i + 1)
        except ValueError:
            break
        else:
            loc_lst.append(loc)
            i = loc
    return loc_lst



# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import pandas as pd
import uuid
import re
from random import randint
import streamlit as st
import py3Dmol
from stmol import showmol
from io import BytesIO

from .file import entry_table_file
from .table import mask_equal, merge_dicts
from .path import (
    load_table,
    load_json,
    get_file_path,
    get_dir_path,
    get_neighbor_path,
    pages_str,
    data_str,
    functions_str,
)
from .lig import lig_col_lst 
from .lst import res_to_lst, str_to_lst,  type_lst
from .color import red_hex
from .col import (rename_col_dict, date_col, nuc_class_col, match_class_col, prot_class_col, 
                gene_class_col, interf_class_col, pocket_class_col, pdb_code_col, chainid_col, ion_lig_col, 
                bound_prot_chainid_col, pharm_lig_col, mem_lig_col, pharm_class_col)

from ..constants.nuc import nuc_class_lst
from ..constants.pharm import match_class_lst, pocket_class_lst, pharm_color_dict, none_pharm_name, other_pharm_name, sp2_name
from ..constants.prot import prot_class_lst, prot_color_dict, none_prot_name, other_prot_name
from ..constants.conf import sw1_name_lst, sw2_name_lst, y32_name_lst, y71_name_lst, y32_name, y71_name, sw1_name, sw2_name, loop_resid_dict, loop_color_dict
from ..constants.gene import gene_class_lst
from ..constants.dimer import interf_class_lst

class_order_dict = {nuc_class_col: nuc_class_lst, match_class_col: match_class_lst, pocket_class_col: pocket_class_lst,
                    prot_class_col: prot_class_lst, sw1_name: sw1_name_lst, sw2_name: sw2_name_lst,
                    y32_name: y32_name_lst, y71_name: y71_name_lst,
                    gene_class_col: gene_class_lst, interf_class_col: interf_class_lst}

sw1_resid_lst = res_to_lst(loop_resid_dict[sw1_name])
sw2_resid_lst = res_to_lst(loop_resid_dict[sw2_name])

mitch_twitter = '<a href="https://twitter.com/Mitch_P?ref_src=twsrc%5Etfw" class="twitter-follow-button" data-show-count="false">Follow @Mitch_P</a><script async src="https://platform.twitter.com/widgets.js" charset="utf-8"></script>'
roland_twitter = '<a href="https://twitter.com/RolandDunbrack?ref_src=twsrc%5Etfw" class="twitter-follow-button" data-show-count="false">Follow @RolandDunbrack</a><script async src="https://platform.twitter.com/widgets.js" charset="utf-8"></script>'


ribbon_name = "Ribbon"
trace_name = "Trace"
standard_name = "Standard"
aa_name = "Amino Acid"

def write_st_end():

    df = load_table(
        get_file_path(
            entry_table_file,
            dir_path=get_neighbor_path(__file__, functions_str, data_str),
        )
    )

    df[date_col] = pd.to_datetime(df[date_col])
    df[date_col] = df[date_col].dt.strftime("%Y-%m")

    st.markdown("---")
    st.markdown(
        "Developed and Maintained by Mitchell Parker, Bulat Faezov, and Roland Dunbrack"
    )
    st.markdown(
        "[Dunbrack Lab](https://dunbrack.fccc.edu/retro/) - [Fox Chase Cancer Center](https://www.foxchase.org)"
    )
    st.markdown(f"Most Recently Deposited Entry {df[date_col].max()}")
    st.markdown("Copyright (c) 2022 Mitchell Isaac Parker")

def reorder_st_cols(df, row, col):

    class_lst = list(class_order_dict.keys())

    if row in class_lst:
        row_order = [x for x in class_order_dict[row] + ["All"] if x in list(df.index.values)] 

    if col in class_lst:
        col_order = [x for x in class_order_dict[col] + ["All"] if x in list(df.columns)] 

    if row in class_lst and col in class_lst:
        df = df.loc[row_order, col_order]
    elif row not in class_lst and col in class_lst:
        df = df.loc[:, col_order]
    elif row in class_lst and col not in class_lst:
        df = df.loc[row_order, :]
    
    return df

def get_st_file_path(st_file):

    return get_file_path(
        f"{randint(0,3261994)}_{st_file.name}",
        dir_path=get_neighbor_path(__file__, functions_str, data_str),
    )


def save_st_file(st_file):

    st_file_path = get_st_file_path(st_file)
    with open(st_file_path, "wb") as file:
        file.write(st_file.getbuffer())
    return st_file_path

def show_st_fig(fig, st_col=None):

    byt = BytesIO()
    fig.savefig(byt, format="png")
    if st_col is None:
        st.image(byt)
    else:
        st_col.image(byt)


def get_html_text(text_color_dict, font_size="medium", font_weight="normal"):

    html_str = ""
    for text, color in text_color_dict.items():

        size = font_size
        if type(font_size) == dict:
            size = font_size[text]

        weight = font_weight
        if type(font_weight) == dict:
            weight = font_weight[text]

        html_str += f'<span style="font-family:sans-serif; font-size: {size}; font-weight: {weight}; color:{color};">{text}</span>'

    return html_str


def load_st_table(file_path, file_name=None, json_format=False):

    if file_name is None:
        file_name = entry_table_file

    file_path = get_file_path(
            file_name,
            dir_path=get_dir_path(
                dir_path=get_neighbor_path(file_path, pages_str, data_str)
            ))

    if json_format:
        return load_json(file_path)
    else:
        return load_table(file_path)
   
def mask_st_table(df, col_dict):

    mask_df = df.copy()

    for col in list(col_dict.keys()):
        val = type_lst(col_dict[col])
        if len(val) > 0:
            if "All" not in val:
                mask_df = mask_equal(mask_df, col, val)

    return mask_df


def rename_st_cols(df, col_lst=None):

    if col_lst is None:
        col_lst = [x for x in list(rename_col_dict.keys()) if x in list(df.columns)]

    return df.loc[:, col_lst].rename(columns=rename_col_dict)


def show_st_dataframe(df, st_col=None, hide_index=True):

    if hide_index:
        hide_dataframe_row_index = """
                <style>
                .row_heading.level0 {display:none}
                .blank {display:none}
                </style>
                """

        st.markdown(hide_dataframe_row_index, unsafe_allow_html=True)

    if st_col is None:
        st.dataframe(df)
    else:
        st_col.dataframe(df)


def show_st_table(df, st_col=None, hide_index=True):

    if hide_index:
        hide_table_row_index = """
                <style>
                tbody th {display:none}
                .blank {display:none}
                </style>
                """
        st.markdown(hide_table_row_index, unsafe_allow_html=True)

    if st_col is None:
        st.table(df)
    else:
        st_col.table(df)


def create_st_button(link_text, link_url, hover_color="#e78ac3", st_col=None):

    button_uuid = str(uuid.uuid4()).replace("-", "")
    button_id = re.sub("\d+", "", button_uuid)

    button_css = f"""
        <style>
            #{button_id} {{
                background-color: rgb(255, 255, 255);
                color: rgb(38, 39, 48);
                padding: 0.25em 0.38em;
                position: relative;
                text-decoration: none;
                border-radius: 4px;
                border-width: 1px;
                border-style: solid;
                border-color: rgb(230, 234, 241);
                border-image: initial;

            }}
            #{button_id}:hover {{
                border-color: {hover_color};
                color: {hover_color};
            }}
            #{button_id}:active {{
                box-shadow: none;
                background-color: {hover_color};
                color: white;
                }}
        </style> """

    html_str = f'<a href="{link_url}" target="_blank" id="{button_id}";>{link_text}</a><br></br>'

    if st_col is None:
        st.markdown(button_css + html_str, unsafe_allow_html=True)
    else:
        st_col.markdown(button_css + html_str, unsafe_allow_html=True)


def download_st_file(file_path, file_name, download_text, st_col=None):

    with open(file_path, "rb") as file:
        if st_col is None:
            st.download_button(download_text, file, file_name=file_name)
        else:
            st_col.download_button(download_text, file, file_name=file_name)


def encode_st_df(df):

    return df.to_csv(sep="\t", index=False).encode("utf-8")


def download_st_df(df, file_name, download_text, st_col=None):

    if st_col is None:
        st.download_button(
            label=download_text,
            data=encode_st_df(df),
            file_name=file_name,
        )
    else:
        st_col.download_button(
            label=download_text,
            data=encode_st_df(df),
            file_name=file_name,
        )


def show_st_3dmol(
    pdb_code,
    style_lst=None,
    label_lst=None,
    reslabel_lst=None,
    zoom_dict=None,
    surface_lst=None,
    cartoon_style="trace",
    cartoon_radius=0.2,
    cartoon_color="lightgray",
    zoom=1,
    spin_on=False,
    width=900,
    height=600,
):

    view = py3Dmol.view(query=f"pdb:{pdb_code.lower()}", width=width, height=height)

    view.setStyle(
        {
            "cartoon": {
                "style": cartoon_style,
                "color": cartoon_color,
                "thickness": cartoon_radius,
            }
        }
    )

    if surface_lst is not None:
        for surface in surface_lst:
            view.addSurface(py3Dmol.VDW, surface[0], surface[1])

    if style_lst is not None:
        for style in style_lst:
            view.addStyle(
                style[0],
                style[1],
            )

    if label_lst is not None:
        for label in label_lst:
            view.addLabel(label[0], label[1], label[2])

    if reslabel_lst is not None:
        for reslabel in reslabel_lst:
            view.addResLabels(reslabel[0], reslabel[1])


    if zoom_dict is None:
        view.zoomTo()
    else:
        view.zoomTo(zoom_dict)

    view.spin(spin_on)

    view.zoom(zoom)
    showmol(view, height=height, width=width)


def show_st_structure(df,
                    mut_resids=None, stick_resids=None,
                    label_muts=None, label_resids=False, label_ligs=False, label_prots=False, 
                    cartoon_style="oval",
                    cartoon_trans=1, surface_trans=0, mut_trans=0.5,
                    mut_color=None, sw1_color=None, sw2_color=None,
                    aa_scheme=False,
                    spin_on=False,
                    all_chains=False,
                    zoom_resids=None,
                    zoom=1.5,
                    width=400,
                    height=400,
                    show_legend=True,
                    legend_font_size="medium",
                    st_col=None):

    if sw1_color is None:
        sw1_color = loop_color_dict[sw1_name]
    if sw2_color is None:
        sw2_color = loop_color_dict[sw2_name]
    if mut_color is None:
        mut_color = red_hex

    stick_resid_lst = res_to_lst(stick_resids)
    mut_resid_lst = res_to_lst(mut_resids)

    style_lst = list()
    label_lst = list()
    reslabel_lst = list()
    surface_lst = list()

    opacity = 0
    if all_chains:
        opacity = 0.75
    
    pdb_code = df[pdb_code_col].iloc[0]
    chainid = df[chainid_col].iloc[0]

    style_lst.append(
        [
            {
                "chain": chainid,
                "invert": True,
            },
            {
                "cartoon": {
                    "color": "white",
                    "style": cartoon_style,
                    "thickness": 0.2,
                    "opacity": opacity,
                }
            },
        ]
    )

    style_lst.append(
        [
            {
                "chain": chainid,
            },
            {
                "cartoon": {
                    "color": "white",
                    "style": cartoon_style,
                    "thickness": 0.2,
                    "opacity": cartoon_trans,
                }
            },
        ]
    )

    surface_lst = [
        [
            {"opacity": surface_trans, "color": "white"},
            {"chain": chainid, "hetflag": False},
        ]
    ]

    if stick_resids is not None:
        if aa_scheme:
            style_lst.append(
                [
                    {"chain": chainid,
                    "resi": stick_resid_lst,
                    "elem": "C"},
                    {"stick": {"colorscheme": "amino", "radius": 0.2}},
                ]
            ) 
            style_lst.append(
                    [
                        {"chain": chainid, "resi": stick_resid_lst},
                        {"stick": {"radius": 0.2}},
                    ]
                )
            surface_lst.append(
            [
                {"opacity": surface_trans,"colorscheme": "amino"},
                {"chain": chainid, "resi": stick_resid_lst, "hetflag": False},
            ]
            )
        else:
            for stick_resid in stick_resid_lst:

                resid_color = "white"
                if int(stick_resid) in sw1_resid_lst:
                    resid_color = sw1_color
                if int(stick_resid) in sw2_resid_lst:
                    resid_color = sw2_color

                style_lst.append(
                [
                    {
                        "chain": chainid,
                        "resi": stick_resid,
                        "elem": "C",
                    },
                    {"stick": {"color": resid_color, "radius": 0.2}},
                ]
            )

                style_lst.append(
                    [
                        {"chain": chainid, "resi": stick_resid},
                        {"stick": {"radius": 0.2}},
                    ]
                )  

        for stick_resid in stick_resid_lst:

            add_reslabel = label_resids
            if type(label_resids) == dict:
                if stick_resid in list(label_resids.keys()):
                    add_reslabel = label_resids[stick_resid]

            if add_reslabel:   
                reslabel_lst.append(
                        [
                            {"chain": chainid,
                            "resi": stick_resid},
                            {
                                "backgroundColor": "lightgray",
                                "fontColor": "black",
                                "backgroundOpacity": 0.5,
                            },
                        ]
                    )


    if mut_resids is not None:
        for mut_resid in mut_resid_lst:
            style_lst.append(
            [
                {
                    "chain": chainid,
                    "resi": mut_resid,
                    "elem": "C",
                },
                {"stick": {"color": mut_color, "radius": 0.2}},
            ]
        )

            style_lst.append(
                [
                    {"chain": chainid, "resi": mut_resid},
                    {"stick": {"radius": 0.2}},
                ]
            )     

            surface_lst.append([
                    
                        {"opacity": mut_trans, "color": mut_color},
                        {"chain": chainid, "resi": mut_resid},
                    
                ])

            add_reslabel = label_muts
            if type(label_muts) == dict:
                if mut_resid in list(label_muts.keys()):
                    add_reslabel = label_muts[mut_resid]

            if add_reslabel:
                reslabel_lst.append(
                        [
                            {"chain": chainid,
                            "resi": mut_resid},
                            {
                                "backgroundColor": "lightgray",
                                "fontColor": "black",
                                "backgroundOpacity": 0.5,
                            },
                        ]
                    )
            
    if not aa_scheme:
        for loop_name, loop_resids in loop_resid_dict.items():

            if loop_name == sw1_name:
                loop_color = sw1_color
            elif loop_name == sw2_name:
                loop_color = sw2_color

            surface_lst.append(
                [
                    {"opacity": surface_trans, "color": loop_color},
                    {"chain": chainid, "resi": loop_resids, "hetflag": False},
                ]
            )

            style_lst.append(
                [
                    {
                        "chain": chainid,
                        "resi": loop_resids,
                    },
                    {
                        "cartoon": {
                            "style": cartoon_style,
                            "color": loop_color,
                            "thickness": 0.2,
                            "opacity": cartoon_trans,
                        }
                    },
                ]
            )

    for lig_col in lig_col_lst:

        lig_lst = str_to_lst(df[lig_col].iloc[0])

        if "None" not in lig_lst:

            lig_style = "stick"
            lig_color = "whiteCarbon"
            lig_scheme = "colorscheme"
            lig_radius = 0.2

            if lig_col == ion_lig_col:

                lig_style = "sphere"
                lig_scheme = "color"
                lig_color = "chartreuse"
                lig_radius = 0.8

            for lig in lig_lst:

                if lig_col != pharm_lig_col:

                    lig_sele = {
                                "resn": lig,
                            }

                    if lig_col != mem_lig_col:
                         lig_sele["chain"] = chainid

                    style_lst.append(
                        [
                            lig_sele,
                            {
                                lig_style: {
                                    lig_scheme: lig_color,
                                    "radius": lig_radius,
                                }
                            },
                        ]
                    )

                add_reslabel = label_ligs
                if type(label_ligs) == dict:
                    if lig in list(label_ligs.keys()):
                        add_reslabel = label_ligs[lig]
                
                if add_reslabel:
                    reslabel_lst.append(
                        [
                            {
                                "chain": chainid,
                                "resn": lig,
                            },
                            {
                                "backgroundColor": "lightgray",
                                "fontColor": "black",
                                "backgroundOpacity": 0.5,
                            },
                        ]
                    )

    pharm_class_lst = str_to_lst(df[pharm_class_col].iloc[0])

    if none_pharm_name not in pharm_class_lst:
        if len(pharm_class_lst) > 1:
            pharm_class = other_pharm_name
        else:
            pharm_class = pharm_class_lst[0]

        pharm_color = pharm_color_dict[pharm_class]

        pharm_lig_lst = str_to_lst(df.at[0, pharm_lig_col])

        for pharm_lig in pharm_lig_lst:
            style_lst.append(
                [
                    {
                        "chain": chainid,
                        "resn": pharm_lig,
                        "elem": "C",
                    },
                    {"stick": {"color": pharm_color, "radius": 0.2}},
                ]
            )
            style_lst.append(
                [
                    {
                        "chain": chainid,
                        "resn": [pharm_lig],
                    },
                    {"stick": {"radius": 0.2}},
                ]
            )

            if pharm_class == sp2_name:
                style_lst.append(
                    [
                        {
                            "chain": chainid,
                            "resi": 12,
                        },
                        {"stick": {"radius": 0.2}},
                    ]
                )

    prot_class_lst = str_to_lst(df[prot_class_col].iloc[0])

    if none_prot_name not in prot_class_lst:
        if len(prot_class_lst) > 1:
            prot_class = other_prot_name
        else:
            prot_class = prot_class_lst[0]
        prot_color = prot_color_dict[prot_class]

        bound_chainid_lst = str_to_lst(df[bound_prot_chainid_col].iloc[0])

        for bound_prot_chainid in bound_chainid_lst:
            style_lst.append(
                [
                    {
                        "chain": bound_prot_chainid,
                    },
                    {
                        "cartoon": {
                            "style": cartoon_style,
                            "color": prot_color,
                            "thickness": 0.2,
                            "opacity": 1,
                        }
                    },
                ]
            )

            add_chainlabel = label_prots
            if type(label_prots) == dict:
                if bound_prot_chainid in list(label_prots.keys()):
                    add_chainlabel = label_prots[bound_prot_chainid]

            if add_chainlabel:

                prot_label = f"Chain {bound_prot_chainid}"

                label_lst.append(
                    [
                        prot_label,
                        {
                            "backgroundColor": "lightgray",
                            "fontColor": "black",
                            "backgroundOpacity": 0.8,
                        },
                        {"chain": bound_prot_chainid},
                    ]
                )

    zoom_dict = {"chain": chainid}

    if zoom_resids is not None:
        if type(zoom_resids) == dict:
            zoom_dict = merge_dicts([zoom_dict, zoom_resids])
        else:
            zoom_dict["resi"] = zoom_resids

    if st_col is None:
        show_st_3dmol(
            pdb_code,
            style_lst=style_lst,
            surface_lst=surface_lst,
            reslabel_lst=reslabel_lst,
            label_lst=label_lst,
            cartoon_style=cartoon_style,
            spin_on=spin_on,
            zoom_dict=zoom_dict,
            zoom=zoom,
            width=width,
            height=height,
        )

    else:
        with st_col:
            show_st_3dmol(
                pdb_code,
                style_lst=style_lst,
                surface_lst=surface_lst,
                reslabel_lst=reslabel_lst,
                label_lst=label_lst,
                cartoon_style=cartoon_style,
                spin_on=spin_on,
                zoom_dict=zoom_dict,
                zoom=zoom,
                width=width,
                height=height,
            )

    if show_legend:
        if aa_scheme:
            aa_type_dict = {
                    'Acidic':['ASP','GLU'],
                    'Basic':['LYS', 'ARG','HIS'],
                    'Polar':['ASN','GLN','SER','THR', 'CYS'],
                    'Nonpolar':['ILE', 'VAL', 'LEU', 'MET','PRO', 'GLY', 'ALA'],
                    'Aromatic':['PHE', 'TYR', 'TRP'],
                    }

            aa_color_dict = {'ASP':'#E60A0A','GLU':'#E60A0A','CYS':'#E6E600','MET':'#E6E600',
                            'LYS':'#145AFF','ARG':'#145AFF','SER':'#FA9600','THR':'#FA9600',
                            'PHE':'#3232AA','TYR':'#3232AA','ASN':'#00DCDC','GLN':'#00DCDC',
                            'GLY':'#C8C8C8','LEU':'#0F820F','VAL':'#0F820F','ILE':'#0F820F',
                            'ALA':'#C8C8C8','TRP':'#B45AB4','HIS':'#8282D2','PRO':'#DC9682'}

            for aa_type, aa_lst in aa_type_dict.items():

                legend_str = get_html_text({f"{aa_type}: ": "#31333F"}, font_weight='bold', font_size=legend_font_size)

                for i, aa_name in enumerate(aa_lst): 

                    legend_str += get_html_text({aa_name:aa_color_dict[aa_name]}, font_size=legend_font_size)

                    if i != len(aa_lst) - 1:
                        legend_str += get_html_text({", ":"#31333F"}, font_size=legend_font_size)

                if st_col is None:
                    st.markdown(legend_str, unsafe_allow_html=True)
                else:
                    st_col.markdown(legend_str, unsafe_allow_html=True)
       
            
            


     




    



# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import pandas as pd
from tqdm import tqdm

from ..functions.table import merge_tables
from ..functions.lst import res_to_lst
from ..functions.coord import build_add_resid_lst
from ..functions.path import save_table
from ..functions.col import (
    core_path_col,
    pdb_code_col,
    chainid_col,
    modelid_col,
    resid_col,
    atomid_col,
    edia_col,
    b_factor_col,
)


def get_index_edia(
    df,
    index,
    edia_dict,
    resid_lst,
    max_ca_dist=4,
    ext_mult=1,
    coord_path_col=None,
):

    if coord_path_col is None:
        coord_path_col = core_path_col

    index_df = pd.DataFrame()

    pdb_code = df.at[index, pdb_code_col]

    if pdb_code in list(edia_dict.keys()):
        chainid = df.at[index, chainid_col]
        if chainid in list(edia_dict[pdb_code].keys()):
            coord_path = df.at[index, coord_path_col]
            modelid = int(df.at[index, modelid_col])

            edia_resid_lst = list(edia_dict[pdb_code][chainid].keys())

            add_resid_lst = build_add_resid_lst(
                coord_path,
                modelid,
                chainid,
                resid_lst,
                edia_resid_lst,
                max_ca_dist=max_ca_dist,
                ext_mult=ext_mult,
            )

            i = 0

            for n, add_resid in enumerate(add_resid_lst):

                if str(add_resid) in edia_resid_lst:

                    for atomid in list(
                        edia_dict[pdb_code][chainid][str(add_resid)].keys()
                    ):

                        index_df.at[i, pdb_code_col] = pdb_code
                        index_df.at[i, chainid_col] = chainid
                        index_df.at[i, resid_col] = n + 1
                        index_df.at[i, atomid_col] = atomid
                        index_df.at[i, edia_col] = edia_dict[pdb_code][chainid][
                            str(add_resid)
                        ][atomid][edia_col]
                        index_df.at[i, b_factor_col] = edia_dict[pdb_code][chainid][
                            str(add_resid)
                        ][atomid][b_factor_col]

                        i += 1

    return index_df


def build_edia_table(
    df, edia_dict, edia_resids, edia_table_path=None, coord_path_col=None
):

    resid_lst = res_to_lst(edia_resids)

    edia_df = pd.DataFrame()

    for index in tqdm(
        list(df.index.values), desc="Building EDIA table", position=0, leave=True
    ):

        edia_df = pd.concat(
            [
                edia_df,
                get_index_edia(
                    df, index, edia_dict, resid_lst, coord_path_col=coord_path_col
                ),
            ],
            sort=False,
        )

    edia_df = merge_tables(left_df=edia_df, right_df=df)

    edia_df = edia_df.reset_index(drop=True)

    if edia_table_path is not None:
        save_table(edia_table_path, edia_df)
    else:
        return edia_df

    print("Built EDIA table!")



# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
import matplotlib.ticker as mticker
from statannot import add_stat_annotation
import seaborn as sns

import warnings

warnings.filterwarnings("ignore", module="seaborn")

from ..functions.lst import type_lst, format_nested_lst
from ..functions.color import change_hex_alpha, get_lst_colors, gray_hex
from ..functions.col import (
    reformat_col_lst,
    resid_col,
    bb_col_lst,
    sc_col_lst,
    phi_col,
    psi_col,
)
from ..functions.table import (
    reformat_val_table,
    lst_col,
    mask_unequal,
    title_str,
    get_ncols,
)
from ..functions.plot import prep_plot_col
from ..functions.path import append_file_path

grid_hex = change_hex_alpha(gray_hex, 0.25)

bb_lim = (-180, 180)
bb_ticks = [-90, 0, 90]
sc_lim = (0, 360)
sc_ticks = [120, 240]

bb_grid_lst = [
    [[-180, 0], [50, 50], [180, 180]],
    [[-180, 0], [-180, -180], [-100, -100]],
    [[-180, 0], [-100, -100], [50, 50]],
    [[0, 180], [-180, -180], [-50, -50]],
    [[0, 180], [100, 100], [180, 180]],
    [[0, 180], [-50, -50], [100, 100]],
]

sc_x_lst = [[[0, 120]], [[120, 240]], [[240, 360]]]

sc_y_lst = [[[0, 0], [120, 120]], [[120, 120], [240, 240]], [[240, 240], [360, 360]]]

sc_grid_lst = list()
for sc_x in sc_x_lst:
    for sc_y in sc_y_lst:
        sc_grid_lst.append(sc_x + sc_y)


def make_facet_plot(
    plot_df,
    x_col,
    plot_path=None,
    y_col=None,
    rename_x=None,
    x_order=None,
    x_palette=None,
    x_count=True,
    hue_col=None,
    rename_hue=None,
    hue_order=None,
    hue_palette=None,
    hue_count=True,
    row_col=None,
    rename_row=None,
    row_order=None,
    row_count=True,
    row_palette=None,
    col_col=None,
    rename_col=None,
    col_order=None,
    col_count=False,
    col_wrap=None,
    col_palette=None,
    plot_width=4,
    plot_height=None,
    font_size=7,
    marker_size=1,
    line_width=0.5,
    tick_len=2,
    show_legend=False,
    legend_pad=5,
    legend_cols=None,
    legend_marker_size=None,
    color_legend_text=False,
    highlight_palette=None,
    highlight_lst=None,
    darken_lst=None,
    darken_palette=None,
    x_lim=None,
    y_lim=None,
    x_ticks=None,
    y_ticks=None,
    x_round=0,
    y_round=0,
    x_rotation=0,
    y_rotation=0,
    x_ha="center",
    y_ha="right",
    x_str=None,
    y_str=None,
    x_pad=None,
    y_pad=None,
    h_lines=None,
    v_lines=None,
    h_color=None,
    v_color=None,
    count_chain=True,
    count_pdb=False,
    count_cf=False,
    plot_scatter=True,
    plot_line=False,
    plot_reg=False,
    plot_kde=False,
    plot_kind=None,
    log_reg=False,
    reg_estimator=None,
    reg_bins=None,
    trun_reg=False,
    kde_common_norm=True,
    kde_common_grid=False,
    kde_bw=1.0,
    kde_thresh=0.05,
    kde_levels=10,
    kde_cut=3.0,
    kde_alpha=0.1,
    rug_alpha=0.5,
    rug_height=0.05,
    rug_width=0.05,
    stat_pairs=None,
    stat_test="t-test_ind",
    stat_loc="inside",
    stat_format="star",
    tick_mult=0.75,
    all_ticks=False,
):

    df = plot_df.copy(deep=True)

    df_col_lst = list(df.columns)

    reformat_lst = list()
    if x_col in reformat_col_lst:
        if x_col not in df_col_lst:
            reformat_lst.append(x_col)
    if y_col in reformat_col_lst:
        if y_col is not None:
            if y_col not in df_col_lst:
                reformat_lst.append(y_col)

    if len(reformat_lst) > 0:
        df = reformat_val_table(df, reformat_lst)

    if plot_kind is None:
        df[x_col] = df[x_col].map(float)
    else:
        if type(x_palette) == str:
            x_palette = [x_palette] * len(lst_col(df, x_col))
        df, x_lst, x_palette = prep_plot_col(
            df,
            x_col,
            color_palette=x_palette,
            rename_vals=rename_x,
            order_lst=x_order,
            label_count=x_count,
            count_chain=count_chain,
            count_pdb=count_pdb,
            count_cf=count_cf,
        )

    if y_col is not None:
        df[y_col] = df[y_col].map(float)

    if hue_col is not None:
        df, hue_lst, hue_palette = prep_plot_col(
            df,
            hue_col,
            color_palette=hue_palette,
            rename_vals=rename_hue,
            order_lst=hue_order,
            label_count=hue_count,
            count_chain=count_chain,
            count_pdb=count_pdb,
            count_cf=count_cf,
        )
        if len(hue_lst) == 1:
            if type(show_legend) != dict and show_legend == False:
                show_legend = False
    else:
        hue_lst = type_lst(0)
        if type(show_legend) != dict:
            show_legend = False

    if row_col is not None:
        df, row_lst, row_palette = prep_plot_col(
            df,
            row_col,
            color_palette=row_palette,
            rename_vals=rename_row,
            order_lst=row_order,
            label_count=row_count,
            count_chain=count_chain,
            count_pdb=count_pdb,
            count_cf=count_cf,
        )

        total_row = len(row_lst)

        if row_palette is not None:
            row_color_dict = get_lst_colors(
                row_lst, palette=row_palette, return_dict=True
            )

        if hue_col == row_col:
            if type(show_legend) != dict:
                show_legend = False
    else:
        row_lst = None
        total_row = 1

    title_col = True
    if col_col is None:
        col_col = resid_col
        if resid_col not in list(df.columns):
            df[resid_col] = "None"
            title_col = False

    df, col_lst, col_palette = prep_plot_col(
        df,
        col_col,
        color_palette=col_palette,
        rename_vals=rename_col,
        order_lst=col_order,
        label_count=col_count,
        count_chain=count_chain,
        count_pdb=count_pdb,
        count_cf=count_cf,
    )
    total_col = len(col_lst)

    if col_palette is not None:
        col_color_dict = get_lst_colors(col_lst, palette=col_palette, return_dict=True)

    df = mask_unequal(df, x_col, 999.00)
    if x_str is None:
        x_str = title_str(x_col)
    if x_col in bb_col_lst:
        x_lim = bb_lim
        if x_ticks is not False:
            x_ticks = bb_ticks
    elif x_col in sc_col_lst and x_lim is None:
        x_lim = sc_lim
        if x_ticks is not False:
            x_ticks = sc_ticks
    if y_col is not None:
        df = mask_unequal(df, y_col, 999.00)
        if y_str is None:
            y_str = title_str(y_col)
        if y_col in bb_col_lst:
            y_lim = bb_lim
            if y_ticks is not False:
                y_ticks = bb_ticks
        elif y_col in sc_col_lst and y_lim is None:
            y_lim = sc_lim
            if y_ticks is not False:
                y_ticks = sc_ticks
    else:
        if y_str is None:
            y_str = "Density"

    if x_lim is not None:
        if type(x_lim) != tuple:
            x_lim = tuple(x_lim)
    if y_lim is not None:
        if type(y_lim) != tuple:
            y_lim = tuple(y_lim)

    if plot_kind is None or hue_col is not None:
        hue_color_dict = get_lst_colors(hue_lst, palette=hue_palette, return_dict=True)
    else:
        hue_color_dict = get_lst_colors(x_lst, palette=x_palette, return_dict=True)
    if highlight_lst is not None:
        highlight_lst = format_nested_lst(highlight_lst, return_str=True)
        highlight_color_lst = get_lst_colors(highlight_lst, palette=highlight_palette)

    if darken_lst is not None:
        darken_lst = format_nested_lst(darken_lst, return_str=True)
        darken_color_lst = get_lst_colors(darken_lst, palette=darken_palette)

    sns.set_context("paper")
    sns.set_style("ticks")
    sns.set_palette(list(hue_color_dict.values()))

    bbox_extra_artists = tuple()

    if row_col is None:
        if col_wrap is None:
            col_wrap = total_col
        total_col = col_wrap
        total_row = total_col / col_wrap
        show_margins = False
    else:
        col_wrap = None
        show_margins = True

    if plot_kind is None:

        g = sns.FacetGrid(
            df,
            hue=hue_col,
            hue_order=hue_lst,
            row=row_col,
            row_order=row_lst,
            col=col_col,
            col_order=col_lst,
            col_wrap=col_wrap,
            margin_titles=show_margins,
        )

        if y_col is None:
            g.map(
                sns.distplot,
                x_col,
                kde=True,
                hist=False,
                rug=True,
                kde_kws={
                    "common_norm": kde_common_norm,
                    "common_grid": kde_common_grid,
                    "bw_adjust": kde_bw,
                    "thresh": kde_thresh,
                    "levels": kde_levels,
                    "cut": kde_cut,
                    "alpha": kde_alpha,
                    "fill": True,
                    "linewidth": line_width,
                },
                rug_kws={"lw": rug_width, "alpha": rug_alpha, "height": rug_height},
            )
        else:

            if plot_line:
                g.map(sns.lineplot, x_col, y_col)
            if plot_reg:
                g.map(
                    sns.regplot,
                    x_col,
                    y_col,
                    scatter_kws={
                        "s": marker_size,
                        "linewidth": 0,
                        "alpha": 0.75,
                    },
                    line_kws={"linewidth": line_width},
                    logx=log_reg,
                    x_estimator=reg_estimator,
                    x_bins=reg_bins,
                    truncate=trun_reg,
                )
            if plot_kde:
                g.map(
                    sns.kdeplot,
                    x_col,
                    y_col,
                    common_norm=kde_common_norm,
                    common_grid=kde_common_grid,
                    bw_adjust=kde_bw,
                    cut=kde_cut,
                    thresh=kde_thresh,
                    levels=kde_levels,
                    fill=True,
                    linewidth=0,
                    alpha=0.25,
                )

            if plot_scatter:
                g.map(
                    plt.scatter,
                    x_col,
                    y_col,
                    s=marker_size,
                    linewidth=0,
                    alpha=0.75,
                )
    else:
        if plot_kind == "strip" or plot_kind == "swarm" or plot_kind == "point":
            g = sns.catplot(
                data=df,
                x=x_col,
                y=y_col,
                hue=hue_col,
                row=row_col,
                col=col_col,
                col_wrap=col_wrap,
                order=x_lst,
                hue_order=hue_lst,
                row_order=row_lst,
                col_order=col_lst,
                kind=plot_kind,
                palette=hue_color_dict,
                legend=False,
                sharex=False,
                sharey=True,
                margin_titles=True,
                s=marker_size,
                linewidth=0,
                alpha=0.75,
                dodge=True,
            )
        else:
            if (
                row_col is not None
                or col_col is not resid_col
                or plot_kind == "count"
                or plot_kind == "bar"
            ):

                if plot_kind == "count":
                    cat_width = 0
                else:
                    cat_width = line_width

                g = sns.catplot(
                    data=df,
                    x=x_col,
                    y=y_col,
                    hue=hue_col,
                    row=row_col,
                    col=col_col,
                    col_wrap=col_wrap,
                    order=x_lst,
                    hue_order=hue_lst,
                    row_order=row_lst,
                    col_order=col_lst,
                    kind=plot_kind,
                    palette=hue_color_dict,
                    legend=False,
                    sharex=False,
                    sharey=True,
                    linewidth=cat_width,
                    margin_titles=True,
                )
            else:
                if plot_kind == "box" or plot_kind == "boxen" or plot_kind == "violin":

                    g = sns.catplot(
                        data=df,
                        x=x_col,
                        y=y_col,
                        hue=hue_col,
                        order=x_lst,
                        hue_order=hue_lst,
                        kind=plot_kind,
                        color="white",
                        legend=False,
                        linewidth=line_width,
                        showfliers=False,
                    )

                    ax = sns.stripplot(
                        data=df,
                        x=x_col,
                        y=y_col,
                        hue=hue_col,
                        order=x_lst,
                        hue_order=hue_lst,
                        palette=hue_color_dict,
                        s=marker_size,
                        alpha=0.75,
                        split=True,
                    )
                    if ax.get_legend() is not None:
                        ax.get_legend().remove()

                    if stat_pairs is not None:
                        x_dict = dict()
                        for val in x_lst:
                            x_dict[val.split(" (N=")[0]] = val

                        box_pairs = list()
                        for i, stat in enumerate(stat_pairs):
                            box_pairs.append((x_dict[stat[0]], x_dict[stat[1]]))

                        add_stat_annotation(
                            ax,
                            data=df,
                            x=x_col,
                            y=y_col,
                            order=x_lst,
                            box_pairs=box_pairs,
                            test=stat_test,
                            text_format=stat_format,
                            loc=stat_loc,
                            verbose=2,
                            fontsize=font_size,
                            linewidth=line_width,
                        )

    if plot_height is None:
        plot_height = plot_width * (total_row / total_col)

    g.fig.set_figheight(plot_height)
    g.fig.set_figwidth(plot_width)

    if plot_kind is None:
        if (x_ticks is None or x_ticks is False) and (
            y_ticks is None or y_ticks is False
        ):
            g.set(
                xlim=x_lim,
                ylim=y_lim,
            )
        elif x_ticks is None or x_ticks is False:
            g.set(
                xlim=x_lim,
                ylim=y_lim,
                yticks=y_ticks,
            )
        elif y_ticks is None or y_ticks is False:
            g.set(
                xlim=x_lim,
                ylim=y_lim,
                xticks=x_ticks,
            )
        else:
            g.set(
                xlim=x_lim,
                ylim=y_lim,
                xticks=x_ticks,
                yticks=y_ticks,
            )
    else:
        if y_ticks is not None and y_lim is not None:
            g.set(
                ylim=y_lim,
                yticks=y_ticks,
            )
        elif y_ticks is not None:
            g.set(
                yticks=y_ticks,
            )
        elif y_lim is not None:
            g.set(
                ylim=y_lim,
            )

    g.fig.subplots_adjust(wspace=0, hspace=0)

    g.fig.add_subplot(111, frameon=False)
    plt.tick_params(labelcolor="none", top=False, bottom=False, left=False, right=False)

    x_label = plt.xlabel(x_str, fontsize=font_size, labelpad=x_pad)
    y_label = plt.ylabel(y_str, fontsize=font_size, labelpad=y_pad)

    bbox_extra_artists += (x_label, y_label)

    if show_legend or type(show_legend) == dict:

        if legend_cols is None:
            legend_cols = get_ncols(hue_lst)

        if legend_marker_size is None:
            legend_marker_size = marker_size * 3

        if type(show_legend) == dict:
            handles = [
                plt.plot(
                    [],
                    [],
                    marker="o",
                    ls="",
                    markersize=legend_marker_size,
                    markeredgewidth=0,
                    markerfacecolor=show_legend[hue],
                    label=hue,
                    color=show_legend[hue],
                )[0]
                for hue in list(show_legend.keys())
            ]
        else:
            handles = [
                plt.plot(
                    [],
                    [],
                    marker="o",
                    ls="",
                    markersize=legend_marker_size,
                    markeredgewidth=0,
                    markerfacecolor=hue_color_dict[hue],
                    label=hue,
                    color=hue_color_dict[hue],
                )[0]
                for hue in hue_lst
            ]

        legend_label_color = None
        if color_legend_text:
            legend_label_color = "linecolor"

        legend = g.fig.legend(
            handles=handles,
            fontsize=font_size,
            ncol=legend_cols,
            loc="upper center",
            frameon=False,
            bbox_transform=g.fig.transFigure,
            bbox_to_anchor=(0.5, 0),
            borderaxespad=legend_pad,
            labelcolor=legend_label_color,
        )

        bbox_extra_artists += (legend,)

    g.set_axis_labels("", "")

    tick_index = 0

    total_axes = len(col_lst)
    if row_col is not None:
        total_axes *= len(row_lst)

    ax_lst = col_lst
    if row_col is not None and col_col is None:
        ax_lst = row_lst
    elif row_col is not None and col_col is not None:
        ax_lst = list()
        for row in row_lst:
            for col in col_lst:
                ax_lst.append((row, col))

    if h_lines is not None:
        h_line_lst = type_lst(h_lines)
        if h_color is None:
            h_color = "black"

    if v_lines is not None:
        v_line_lst = type_lst(v_lines)
        if v_color is None:
            v_color = "black"

    for i, ax in enumerate(g.axes.flat):

        ax_name = ax_lst[i]

        if h_lines is not None:
            for h in h_line_lst:
                ax.axhline(y=h, linewidth=line_width, linestyle="--", color=h_color)

        if v_lines is not None:
            for v in v_line_lst:
                ax.axvline(x=v, linewidth=line_width, linestyle="--", color=v_color)

        clean_name = ax_name
        if type(clean_name) == str:
            if " (N=" in clean_name:
                clean_name = clean_name.split(" (N=")[0]
        elif type(clean_name) == tuple:
            clean_name = list(clean_name)
            for n, name in enumerate(clean_name):
                clean_name[n] = name.split(" (N=")[0]
            clean_name = tuple(clean_name)

        ax_grid_hex = grid_hex
        if darken_lst is not None:
            for d, darken in enumerate(darken_lst):
                if clean_name in darken:
                    ax.set_facecolor(darken_color_lst[d])
                    ax_grid_hex = "white"

        if x_col == phi_col and y_col == psi_col:
            for bb_grid in bb_grid_lst:
                ax.fill_between(
                    bb_grid[0],
                    bb_grid[1],
                    bb_grid[2],
                    facecolor="none",
                    edgecolor=ax_grid_hex,
                    linewidth=line_width,
                    zorder=0,
                )
        elif (x_col in sc_col_lst and y_col in sc_col_lst) and (
            x_ticks is None and y_ticks is None
        ):
            for sc_grid in sc_grid_lst:
                ax.fill_between(
                    sc_grid[0],
                    sc_grid[1],
                    sc_grid[2],
                    facecolor="none",
                    edgecolor=ax_grid_hex,
                    linewidth=line_width,
                    zorder=0,
                )
        else:
            if plot_kind is None:
                ax.xaxis.grid(color=ax_grid_hex, linewidth=line_width)
            ax.yaxis.grid(color=ax_grid_hex, linewidth=line_width)

        ax.set_axisbelow(True)

        if row_col is None:
            if ax.get_title():
                ax.set_title("")

            if len(col_lst) > 1:

                highlight_color = gray_hex
                if highlight_lst is not None:
                    for h, highlight in enumerate(highlight_lst):
                        if clean_name in highlight:
                            highlight_color = highlight_color_lst[h]

                divider = make_axes_locatable(ax)
                cax = divider.append_axes("top", size="20%", pad=0)
                cax.get_xaxis().set_visible(False)
                cax.get_yaxis().set_visible(False)
                cax.set_facecolor(highlight_color)
                for _, spine in cax.spines.items():
                    spine.set_visible(True)
                    spine.set_color("black")
                    spine.set_linewidth(line_width)
                cax.text(
                    0.5,
                    0.4,
                    str(ax_name),
                    color="white",
                    fontsize=font_size * 0.75,
                    verticalalignment="center",
                    horizontalalignment="center",
                    transform=cax.transAxes,
                )
        else:
            if ax.get_title():

                if title_col:
                    col = ax.get_title().split("=", 1)[1]
                else:
                    col = ""

                col_color = "black"
                if col_palette is not None:
                    col_color = col_color_dict[col]

                col_txt = ax.set_title(
                    col,
                    fontsize=font_size,
                    color=col_color,
                )

                bbox_extra_artists += (col_txt,)

            if ax.texts:
                row = ax.texts[0].get_text().split("=", 1)[1]

                row_color = "black"
                if row_palette is not None:
                    row_color = row_color_dict[row[1:]]

                row_txt = ax.text(
                    ax.texts[0].get_unitless_position()[0],
                    ax.texts[0].get_unitless_position()[1],
                    row,
                    transform=ax.transAxes,
                    va="center",
                    fontsize=font_size,
                    color=row_color,
                )
                ax.texts[0].remove()

                bbox_extra_artists += (row_txt,)

        add_ticks = False
        bottom_label = True
        left_label = True
        bottom = True
        left = True

        if total_col == 1:
            add_ticks = True
            if i != total_axes - 1:
                bottom_label = False
                bottom = False
        elif total_axes == 1 and row_col is None:
            add_ticks = True
        elif total_axes == col_wrap and row_col is None:
            add_ticks = True
            if i != 0:
                left_label = False
                left = False
        elif i % total_col == 0:
            if tick_index % 2 != 0 or all_ticks:
                add_ticks = True
            tick_index += 1
            if not all_ticks:
                bottom_label = False
                bottom = False
        elif (total_axes - i) < total_col + 1:
            if i % 2 != 0 or all_ticks:
                add_ticks = True
            left_label = False
            left = False
        else:
            bottom_label = False
            left_label = False
            bottom = False
            left = False

        if add_ticks:

            tick_format = "{:,."

            if plot_kind is None:
                x_tick_format = tick_format + str(x_round)

                x_tick_format += "f}"

                x_tick_lst = ax.get_xticks()
                ax.xaxis.set_major_locator(mticker.FixedLocator(x_tick_lst))
                ax.set_xticklabels(
                    [x_tick_format.format(x) for x in x_tick_lst],
                    fontsize=font_size * tick_mult,
                    rotation=x_rotation,
                    ha=x_ha,
                )
            else:
                ax.set_xticklabels(
                    ax.get_xticklabels(),
                    fontsize=font_size * tick_mult,
                    rotation=45,
                    ha="right",
                )

            y_tick_format = tick_format + str(y_round)
            y_tick_format += "f}"

            y_tick_lst = ax.get_yticks()
            ax.yaxis.set_major_locator(mticker.FixedLocator(y_tick_lst))
            ax.set_yticklabels(
                [y_tick_format.format(y) for y in y_tick_lst],
                fontsize=font_size * tick_mult,
                rotation=y_rotation,
                ha=y_ha,
            )

        elif not add_ticks:
            bottom_label = False
            left_label = False

        if x_ticks is False:
            bottom_label = False
            bottom = False
        if y_ticks is False:
            left = False
            left_label = False

        ax.tick_params(
            direction="out",
            labelbottom=bottom_label,
            labelleft=left_label,
            bottom=bottom,
            left=left,
            length=tick_len,
            width=line_width,
            colors="black",
        )

        for _, spine in ax.spines.items():
            spine.set_visible(True)
            spine.set_color("black")
            spine.set_linewidth(line_width)

    if plot_path is None:
        return g
    else:
        append_file_path(plot_path)

        if 'png' in plot_path:
            plot_format = 'png'
        if 'pdf' in plot_path:
            plot_format = 'pdf'

        plt.savefig(
            plot_path,
            format=plot_format,
            bbox_extra_artists=bbox_extra_artists,
            bbox_inches="tight",
            pad_inches=0.0,
            dpi=600,
        )
        plt.close()

        print("Made facet plot!")



# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import math

from .lst import lst_unique


def calc_q_score(i_cont_lst, j_cont_lst, i_dist_lst, j_dist_lst):

    cont_lst = lst_unique(i_cont_lst, j_cont_lst)

    q_sum = 0
    w_sum = 0

    for cont in cont_lst:

        i_cb_dist = 12
        if cont in i_cont_lst:
            i_cb_dist = i_dist_lst[i_cont_lst.index(cont)]

        j_cb_dist = 12
        if cont in j_cont_lst:
            j_cb_dist = j_dist_lst[j_cont_lst.index(cont)]

        min_cb_diff = min(i_cb_dist, j_cb_dist)
        abs_cb_diff = abs(i_cb_dist - j_cb_dist)

        if min_cb_diff > 5:
            weight = math.exp(-((min_cb_diff - 5) ** 2) / 9.158)
        else:
            weight = 1

        q_sum += weight * math.exp(-0.5 * abs_cb_diff)
        w_sum += weight

    if w_sum == 0:
        q_score = 1
    else:
        q_score = 1 - (q_sum / w_sum)

    return q_score


# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

from ..functions.color import gray_hex
from ..scripts.prep_pocket import pocket_unbound_name
from .pharm import sp12_name, sp2_name, other_pharm_name

pocket_site_lst = [sp12_name, sp2_name, other_pharm_name]

pocket_color_dict = {
    sp12_name: "#1b9e77",
    sp2_name: "#d95f02",
    pocket_unbound_name: gray_hex,
}


# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import pandas as pd
import numpy as np

from ..scripts.make_facet_plot import make_facet_plot
from ..functions.color import change_hex_alpha, get_hex, gray_hex
from ..functions.chem import draw_lig_plot, get_lig_mcs, get_lig_smiles, get_lig_simi, is_lig_match
from ..functions.table import build_col_count_dict, mask_equal, lst_col, mask_unequal, str_to_dict, fix_col, make_dict
from ..functions.col import (pocket_class_col, pdb_id_col, pdb_code_col, id_col,
                            bound_lig_cont_col, chainid_col, pharm_lig_smiles_col, pharm_class_col,
                            nuc_class_col, mut_status_col, prot_class_col, interf_class_col,
                            pocket_lig_col, pharm_lig_col, pocket_score_col, match_class_col, 
                            bio_lig_col, modelid_col, gene_class_col, bound_prot_swiss_id_col,
                            pocket_volume_col, pharm_lig_col, rename_col_dict)
from ..functions.file import pocket_table_file, entry_table_file
from ..functions.path import rascore_str
from ..constants.pharm import sp2_name, sp12_name, bound_name, pharm_color_dict
from ..constants.conf import loop_resid_dict, loop_color_dict, sw1_name, sw2_name, y32_name, y71_name
from ..functions.gui import (load_st_table, show_st_dataframe, write_st_end,
                            show_st_structure, show_st_table, get_html_text,
                            download_st_df, rename_st_cols, reorder_st_cols,
                            ribbon_name, trace_name, standard_name, aa_name)
from ..functions.lst import calc_simpson, res_to_lst,lst_unique, str_to_lst, lst_nums

import streamlit as st

reverse_col_dict = make_dict(list(rename_col_dict.values()),list(rename_col_dict.keys()))

sp12_bound_name = f"{sp12_name}-{bound_name}"
sp2_bound_name = f"{sp2_name}-{bound_name}"

mean_name = "Mean"
max_name = "Max"
min_name = "Min"

sw1_resid_lst = res_to_lst(loop_resid_dict[sw1_name])
sw2_resid_lst = res_to_lst(loop_resid_dict[sw2_name])

def inhibitor_page():
    
    st.markdown("## Compare Inhibitors")

    st.markdown("---")

    df = load_st_table(__file__)
    pocket_df = load_st_table(__file__, file_name=pocket_table_file)

    pocket_df = mask_unequal(pocket_df, pocket_lig_col, "STP")
    df = mask_equal(df, pdb_id_col, lst_col(pocket_df, pdb_id_col))

    st.sidebar.markdown("**Note.** Use [PubChem Draw Tool](https://pubchem.ncbi.nlm.nih.gov/#draw=true) to design and edit SMILES strings.")

    pocket_class = st.sidebar.radio(rename_col_dict[pocket_class_col], [sp12_bound_name, sp2_bound_name])
    pharm_class = pocket_class.split('-')[0]

    pharm_df = mask_equal(df, pharm_class_col, pharm_class)

    for col in [pocket_score_col, pocket_volume_col]:

        pharm_df[col] = pharm_df[pdb_id_col].map(make_dict(lst_col(pocket_df, pdb_id_col),lst_col(pocket_df,col)))

    left_name = "Left"
    right_name = "Right"

    a_name = "Group A"
    b_name = "Group B"

    left_query_col, right_query_col = st.columns(2)

    left_name = st.sidebar.text_input(label=f"{left_name} Label", value=a_name)
    right_name = st.sidebar.text_input(label=f"{right_name} Label", value=b_name)

    query_col_dict = {left_name: left_query_col, right_name: right_query_col}

    query_df_dict = dict()

    for query_name, query_col in query_col_dict.items():

        query_df = pharm_df.copy(deep=True)

        query_col.markdown(f"#### Site Search ({query_name})")

        cont_lst = query_col.multiselect(f"Residue Contacts ({query_name})", lst_nums(1, 189))

        if len(cont_lst) > 0:

            cont_simi = query_col.slider(f"Residue Contact Simpson Similarity ({query_name})",  min_value=0.0, max_value=1.0, value=0.4)

            query_col.write(f"Searching for Sites ({query_name})")
            site_bar = query_col.progress(0)
            site_index_lst = list()
            for i, index in enumerate(list(query_df.index.values)):
                site_cont_dict = str_to_dict(query_df.at[index, bound_lig_cont_col], return_int=True)
                site_cont_lst = site_cont_dict[query_df.at[index, pharm_lig_col]]
                if calc_simpson(site_cont_lst, cont_lst) >= cont_simi:
                    site_index_lst.append(index)

                site_bar.progress((i + 1)/len(query_df))

            query_df = query_df.loc[site_index_lst, :]

        if len(query_df) == 0:
            query_col.warning("Insufficient Number of Structures Based On Site Query")
        else:
            query_col.markdown(f"#### Chemistry Search ({query_name})")

            smiles_str = query_col.text_input(f"Comma-Separated SMILES Strings ({query_name})")

            if len(smiles_str) > 0:

                smiles_lst = str_to_lst(smiles_str)

                exact_name = 'Exact'
                simi_name = 'Similarity'

                match_type = query_col.radio(f"Substructure Match ({query_name})", [exact_name, simi_name])

                if match_type == exact_name:
                    min_match = query_col.number_input(f"Minimum Matches ({query_name})", min_value=1, max_value=len(smiles_lst), value=1)

                elif match_type == simi_name:

                    min_simi = query_col.slider(f"2D Fingerprint Similarity ({query_name})", min_value=0.0, max_value=1.0, value=0.1)

                    if len(smiles_lst) > 1:
                        simi_method = query_col.radio(f"Similarity Method ({query_name})", [mean_name, max_name, min_name])

                query_col.write(f"Searching for Chemistries ({query_name})")
                chem_bar = query_col.progress(0.0)
                chem_index_lst = list()
                for i, index in enumerate(list(query_df.index.values)):
                    chem_smiles_dict = str_to_dict(query_df.at[index, pharm_lig_smiles_col], return_str=True)
                    chem_smiles= chem_smiles_dict[query_df.at[index, pharm_lig_col]][0]
                    if match_type == exact_name:
                        if is_lig_match(pharm_lig=chem_smiles,matches=smiles_lst) >= min_match:
                            chem_index_lst.append(index)
                    elif match_type == simi_name:
                        lig_simi_lst = [get_lig_simi(chem_smiles, x) for x in smiles_lst]
                        if len(smiles_lst) == 1:
                            lig_simi = lig_simi_lst[0]
                        else:
                            if simi_method == mean_name:
                                lig_simi = np.mean(lig_simi_lst)
                            elif simi_method == max_name:
                                lig_simi = np.max(lig_simi_lst)
                            elif simi_method == min_name:
                                lig_simi = np.min(lig_simi_lst)

                        if lig_simi >= min_simi:
                            chem_index_lst.append(index)

                    chem_bar.progress((i + 1)/len(query_df))

                    query_df = query_df.loc[chem_index_lst, :]

            if len(query_df) == 0:
                query_col.warning("Insufficient Number of Structures Based On Chemistry Search")
            else:
                query_df_dict[query_name] = query_df
                if query_col.checkbox(f"Display Selection Options ({query_name})", value=False):

                    query_col.markdown(f"#### Conformation Selection ({query_name})")

                    for col in [sw1_name, sw2_name, y32_name, y71_name]:

                        mask_lst = query_col.multiselect(f"{rename_col_dict[col]} ({query_name})", lst_col(query_df, col, unique=True))

                        if len(mask_lst) > 0:
                            query_df = mask_equal(query_df, col, mask_lst)

                    query_col.markdown(f"#### Annotation Selection ({query_name})")

                    for col in [gene_class_col, nuc_class_col, mut_status_col, prot_class_col, match_class_col, interf_class_col]:

                        mask_lst = query_col.multiselect(f"{rename_col_dict[col]} ({query_name})", lst_col(query_df, col, unique=True))

                        if len(mask_lst) > 0:
                            query_df = mask_equal(query_df, col, mask_lst)

                    query_df_dict[query_name] = query_df

    if len(query_df_dict.keys()) == 2:

        left_df = query_df_dict[left_name]
        right_df = query_df_dict[right_name]

        left_lst = lst_col(left_df, pdb_id_col, unique=True)
        right_lst = lst_col(right_df, pdb_id_col, unique=True)

        many_df = pd.concat([left_df, right_df], sort=False)
        many_df = many_df.drop_duplicates()
        many_df = many_df.reset_index(drop=True)

        both_name = "Both Labels"

        for index in list(many_df.index.values):

            pdb_id = many_df.at[index, pdb_id_col]

            if pdb_id in left_lst and pdb_id in right_lst:
                label = both_name
            elif pdb_id in left_lst:
                label = left_name
            elif pdb_id in right_lst:
                label = right_name

            many_df.at[index, id_col] = label

        with st.expander("One-to-One Comparison", expanded=False):

            st.markdown("#### One-to-One")

            left_info_col, right_info_col = st.columns(2)   

            info_col_dict = {left_name: left_info_col, right_name: right_info_col}   

            pdb_name_dict = dict()
            pdb_df_dict = dict()

            for query_name, query_df in query_df_dict.items():

                pdb_id_lst = [x.upper() for x in lst_col(query_df, pdb_id_col) if x.upper() not in list(pdb_name_dict.values())]

                info_col = info_col_dict[query_name]
            
                pdb_upper = info_col.selectbox(f"PDB ID ({query_name})", pdb_id_lst)

                pdb_code = pdb_upper[:4].lower()
                chainid = pdb_upper[4:5]

                pdb_id = f"{pdb_code}{chainid}"
             
                pdb_df = mask_equal(query_df, pdb_id_col, pdb_id)
                
                pdb_name_dict[query_name] = pdb_upper
                pdb_df_dict[query_name] = pdb_df

                pdb_code = pdb_df[pdb_code_col].iloc[0]
                chainid = pdb_df[chainid_col].iloc[0]
                gene_class = pdb_df[gene_class_col].iloc[0]
                mut_status = pdb_df[mut_status_col].iloc[0]
                pharm_lig = pdb_df[pharm_lig_col].iloc[0]
                smiles_dict = str_to_dict(pdb_df[pharm_lig_smiles_col].iloc[0], return_str=True)
                cont_dict = str_to_dict(pdb_df[bound_lig_cont_col].iloc[0], return_int=True)

                smiles = smiles_dict[pharm_lig][0]

                info_col.markdown(f"##### PDB: [{pdb_code.upper()}](https://www.rcsb.org/structure/{pdb_code}) (Chain {chainid}) - {gene_class}({mut_status})")

                col_pair_dict = {nuc_class_col: bio_lig_col, match_class_col: pharm_lig_col, prot_class_col: bound_prot_swiss_id_col}
                for col_1, col_2 in col_pair_dict.items():
                    val_1 = pdb_df[col_1].iloc[0]
                    val_2 = pdb_df[col_2].iloc[0]
                    col_str = f"**{rename_col_dict[col_1]}:** {val_1}"
                    if val_1 != "None":
                        if col_2 in [bio_lig_col, pharm_lig_col] and val_2 != "None":
                            col_str += f" ([{val_2}](https://www.rcsb.org/ligand/{val_2}))"
                        elif col_2 in [bound_prot_swiss_id_col] and val_2 != "None":
                            col_str += f" ([{val_2}](https://www.uniprot.org/uniprot/{val_2}))"
                        else:
                            col_str += f" ({val_2})"
                    info_col.markdown(col_str)

                info_col.markdown(f"**SMILES String:** {smiles}")
                
                for col in [pocket_score_col, pocket_volume_col]:
                    info_col.markdown(f"**{rename_col_dict[col]}:** {round(float(pdb_df[col].iloc[0]),2)}")

            st.markdown("---")

            st.markdown("##### Viewer Settings") 

            left_set_col, right_set_col = st.columns(2)  

            label_resids = left_set_col.checkbox("Label Residues",value=True)

        
            scheme_dict = {standard_name: False, aa_name: True}
            aa_scheme = scheme_dict[
                left_set_col.radio("Residue Color Scheme", [standard_name, aa_name])
            ]

            style_dict = {ribbon_name: "oval", trace_name: "trace"}
            cartoon_style = style_dict[
                left_set_col.radio("Cartoon Style", [ribbon_name, trace_name])
            ]

            cartoon_trans = right_set_col.slider(
                "Cartoon Transparency", min_value=0.0, max_value=1.0, value=0.5
            )

            surface_trans = right_set_col.slider(
                "Surface Transparency", min_value=0.0, max_value=1.0, value=0.7
            )

            left_view_col, right_view_col = st.columns(2)   

            view_col_dict = {left_name: left_view_col, right_name: right_view_col}  

            for query_name, query_df in query_df_dict.items():
                
                pdb_df = pdb_df_dict[query_name] 
                
                pharm_lig = pdb_df[pharm_lig_col].iloc[0]   
                cont_dict = str_to_dict(pdb_df[bound_lig_cont_col].iloc[0], return_int=True)

                view_col = view_col_dict[query_name]

                show_st_structure(pdb_df,
                        stick_resids=cont_dict[pharm_lig],
                        label_resids=label_resids, 
                        cartoon_style=cartoon_style,
                        cartoon_trans=cartoon_trans, 
                        surface_trans=surface_trans,
                        aa_scheme=aa_scheme,
                        zoom_resids={"resn":pharm_lig},
                        zoom=0.7,
                        width=400,
                        height=400,
                        st_col=view_col)

                label_size = "medium"

                if not aa_scheme:

                    for col in [sw1_name, sw2_name, y32_name, y71_name]:

                        label_str = get_html_text({f"{rename_col_dict[col]}: ": "#31333F"}, font_weight='bold', font_size=label_size)

                        if col in [y32_name, sw1_name]:
                            label_color = loop_color_dict[sw1_name]
                        elif col in [y71_name, sw2_name]:
                            label_color = loop_color_dict[sw2_name]

                        label_str += get_html_text({pdb_df[col].iloc[0]: label_color}, font_size=label_size)

                        view_col.markdown(label_str, unsafe_allow_html=True)

            st.markdown("---")

            left_site_col, right_chem_col = st.columns(2)

            pdb_1 = pdb_name_dict[left_name]
            pdb_2 = pdb_name_dict[right_name]

            pdb_1_df = pdb_df_dict[left_name]
            pdb_2_df = pdb_df_dict[right_name]

            lig_1 = pdb_1_df[pharm_lig_col].iloc[0]
            lig_2 = pdb_2_df[pharm_lig_col].iloc[0]

            smiles_dict_1 = str_to_dict(pdb_1_df[pharm_lig_smiles_col].iloc[0])
            smiles_dict_2 = str_to_dict(pdb_2_df[pharm_lig_smiles_col].iloc[0])
            
            smiles_1 = smiles_dict_1[lig_1][0]
            smiles_2 = smiles_dict_2[lig_2][0]

            right_chem_col.markdown("#### Inhibitor Chemistry")

            mcs = get_lig_mcs([smiles_1, smiles_2])

            right_chem_col.markdown(f"**2D Fingerprint Similarity:** {round(get_lig_simi(smiles_1, smiles_2),2)}")

            right_chem_col.markdown(f"**Maximum Common Substructure (MCS):** {get_lig_smiles(mcs)}")

            if not right_chem_col.checkbox("Highlight MCS",value=True):
                mcs = None
            
            right_chem_col.image(draw_lig_plot([smiles_1, smiles_2], 
            lig_labels=[f"{pdb_1} ({lig_1})", f"{pdb_2} ({lig_2})"], font_size=16, mol_pad=0,
            plot_height=4, plot_width=3, highlight_querys=mcs, color_palette=[pharm_color_dict[pharm_class]]))

            cont_dict_1 = str_to_dict(pdb_1_df[bound_lig_cont_col].iloc[0])
            cont_dict_2 = str_to_dict(pdb_2_df[bound_lig_cont_col].iloc[0])

            cont_lst_1 = cont_dict_1[lig_1]
            cont_lst_2 = cont_dict_2[lig_2]
            
            cont_df = pd.DataFrame()

            left_site_col.markdown("#### Inhibitor Site")

            left_site_col.markdown(f"**Residue Contact Similarity (Simpson):** {round(calc_simpson(cont_lst_1, cont_lst_2),2)}")

            cont_col_1 = f"{pdb_1} ({lig_1})"
            cont_col_2 = f"{pdb_2} ({lig_2})"

            for resid in sorted(lst_unique(cont_lst_1 + cont_lst_2,return_int=True)):

                cont_1 = ''
                cont_2 = ''

                if str(resid) in cont_lst_1:
                    cont_1 = "✓"
                if str(resid) in cont_lst_2:
                    cont_2 = "✓"

                cont_df.at[resid, cont_col_1] = cont_1
                cont_df.at[resid, cont_col_2] = cont_2

            cont_df = cont_df.rename_axis('Residue Contact').reset_index()

            show_st_table(cont_df, st_col=left_site_col)


        with st.expander("Many-to-Many Comparison", expanded=False):

            st.markdown("#### Many-to-Many Comparison")

            count_dict = build_col_count_dict(many_df, id_col)

            color_dict = {left_name: pharm_color_dict[pharm_class], 
            right_name: get_hex(change_hex_alpha(pharm_color_dict[pharm_class], 0.5)), both_name: gray_hex}

            left_plot_col, middle_plot_col, right_plot_col = st.columns(3)

            for label_name, label_color in color_dict.items():

                count = 0
                if label_name in list(count_dict.keys()):
                    count = count_dict[label_name]

                color_dict[label_name] = left_plot_col.color_picker(label=f"{label_name} (N={count})", 
                                                                value=label_color)

            many_order =  [x for x in list(color_dict.keys()) if x in lst_col(many_df, id_col, unique=True)]

            scatter_name = "Scatterplot"
            box_name = "Boxplot"

            plot_type = middle_plot_col.radio("Plot Type", [scatter_name, box_name])

            marker_size = middle_plot_col.number_input("Marker Size", min_value=1, max_value=50, value=5)
            
            if plot_type == scatter_name:
                plot_reg = middle_plot_col.checkbox("Display Regression",value=True)

                fig = make_facet_plot(many_df, 
                    x_col=pocket_volume_col,
                    y_col=pocket_score_col,
                    x_str=rename_col_dict[pocket_volume_col],
                    y_str=rename_col_dict[pocket_score_col],
                    x_round=0,
                    y_round=1,
                    hue_col=id_col,
                    hue_palette=color_dict,
                    hue_order=many_order,
                    plot_width=3,
                    plot_height=3,
                    font_size=12,
                    x_rotation=45,
                    marker_size=marker_size,
                    x_pad=25,
                    y_pad=10,
                    x_ha="right",
                    plot_reg=plot_reg,
                    log_reg=True,
                    trun_reg=True,
                    x_lim=[0, 1700],
                    x_ticks=[0, 500, 1000, 1500],
                    y_lim=[0, 1.2],
                    y_ticks=[0.0, 0.5, 1.0])
            
            elif plot_type == box_name:

                y_str = middle_plot_col.radio("Y-Axis",[rename_col_dict[pocket_score_col],rename_col_dict[pocket_volume_col]])

                y_col = reverse_col_dict[y_str]

                if y_col == pocket_score_col:
                    y_lim = [0, 1.2]
                    y_ticks = [0.0, 0.5, 1.0]
                    y_round = 1

                elif y_col == pocket_volume_col:
                    y_lim = [0, 1700]
                    y_ticks = [0, 500, 1000, 1500]
                    y_round = 0

                stat_pairs = None
                if len(many_order) > 1:
                    if middle_plot_col.checkbox("Display Welch's t-test p-value",value=True):
                        stat_pairs = list()
                        for id_1 in many_order:
                            for id_2 in many_order:
                                if id_1 != id_2:
                                    if (id_2, id_1) not in stat_pairs:
                                        stat_pairs.append((id_1, id_2))

                fig = make_facet_plot(
                        many_df,
                        x_col=id_col,
                        x_order=many_order,
                        x_palette=color_dict,
                        x_count=True,
                        x_str="",
                        y_col=y_col,
                        y_str=y_str,
                        show_legend=False,
                        plot_width=3,
                        plot_height=3,
                        plot_kind="box",
                        font_size=12,
                        marker_size=marker_size,
                        x_pad=35,
                        y_lim=y_lim,
                        y_ticks=y_ticks,
                        y_round=y_round,
                        stat_pairs=stat_pairs,
                        stat_test="t-test_welch",
                        stat_format="simple"
                
                    
                    )

            right_plot_col.pyplot(fig)

            many_df[pocket_score_col] = many_df[pocket_score_col].map(float)
            many_df[pocket_volume_col] = many_df[pocket_volume_col].map(float)

            left_stat_col, right_stat_col = st.columns(2)

            stat_col_dict = {pocket_score_col: left_stat_col, pocket_volume_col: right_stat_col}

            for stat_name, stat_col in stat_col_dict.items():

                stat_col.markdown(f"##### {rename_col_dict[stat_name]}")

                temp_df = pd.pivot_table(many_df, values=stat_name,
                                        index=id_col,
                                        aggfunc=[np.mean, min, max])

                stat_df = pd.DataFrame()
                for index in list(temp_df.index.values):
                    for col, stat in zip(list(temp_df.columns), [mean_name, min_name, max_name]):
                        stat_df.at[index, stat] = temp_df.at[index, col]

                for col in list(stat_df.columns):
                    stat_df[col] = stat_df[col].round(1)
                    stat_df[col] = stat_df[col].map(str)
                    stat_df = fix_col(stat_df, col)

                stat_df = stat_df.loc[many_order, :]

                show_st_table(stat_df.rename_axis(rename_col_dict[id_col]).reset_index(), st_col=stat_col)

            st.markdown("---")

            cont_df = pd.DataFrame()

            cont_order = list()

            for index in list(many_df.index.values):
                
                pharm_lig = many_df.at[index, pharm_lig_col]
                cont_dict = str_to_dict(many_df.at[index, bound_lig_cont_col], return_int=True)

                cont_lst = cont_dict[pharm_lig]

                label = many_df.at[index, id_col]
                label_name = f"{label} (N={count_dict[label]})"

                if label_name not in list(cont_df.columns):
                    cont_df[label_name] = 0

                for cont in cont_lst:
                    if cont in cont_order:
                        cont_df.at[cont, label_name] += 1
                    else:
                        cont_df.at[cont, label_name] = 1
                        cont_order.append(cont)

            cont_df = cont_df.fillna(0)

            for col in list(cont_df.columns):
                cont_df[col] = (cont_df[col] / count_dict[col.split(" (")[0]]) * 100
                cont_df[col] = cont_df[col].round(1)
                cont_df[col] = cont_df[col].map(str)
                cont_df = fix_col(cont_df, col)
                cont_df[col] += "%"
                cont_df[col] = cont_df[col].replace("0%","")

            cont_df = cont_df.loc[sorted(cont_order), [f"{x} (N={count_dict[x]})" for x in many_order if x in list(count_dict.keys())]]

            cont_df = cont_df.rename_axis('Residue Contact').reset_index()

            left_table_col, right_table_col = st.columns(2)

            right_table_col.markdown("##### Inhibitor Site")

            show_st_table(cont_df, st_col=right_table_col)

            for table_col in [y32_name, y71_name, sw1_name, sw2_name]:

                left_table_col.markdown(f"##### {rename_col_dict[table_col]}")

                loop_df = (
                    pd.pivot_table(
                        data=rename_st_cols(many_df),
                        index=[rename_col_dict[table_col]],
                        columns=rename_col_dict[id_col],
                        values=rename_col_dict[pdb_id_col],
                        aggfunc="nunique",
                        margins=True,
                    )
                    .fillna("")
                )

                for col in list(loop_df.columns):
                    loop_df[col] = loop_df[col].map(str)
                    loop_df = fix_col(loop_df, col)

                loop_df = reorder_st_cols(loop_df, table_col, id_col)

                loop_df = loop_df.reset_index()

                show_st_table(loop_df, st_col=left_table_col)


        with st.expander("Entries Table", expanded=False):

            st.markdown(f"##### Entries Table")

            show_df = rename_st_cols(many_df)

            del show_df[rename_col_dict[pdb_id_col]]
            del show_df[rename_col_dict[modelid_col]]

            show_st_dataframe(show_df)

            entries_file_name = st.text_input(
                label="Entries File Name",
                value=f"{rascore_str}_{entry_table_file}",
            )
            download_st_df(show_df, entries_file_name, "Download Entries Table")            

    write_st_end()


from src.download.modules import *


def what_is_left_to_download(files_were_found, file_names_with_null_if_files_absent):
    left_to_download = list()
    df_file_names_with_null_if_files_absent = pd.DataFrame(file_names_with_null_if_files_absent, columns=['all'])
    df_files_were_found = pd.DataFrame(files_were_found, columns=['found'])

    merged_all_vs_found = df_file_names_with_null_if_files_absent.merge(df_files_were_found, how="left", left_on="all", right_on="found")
    difference_all_vs_found = merged_all_vs_found.where((merged_all_vs_found["found"] != merged_all_vs_found["all"]))

    for file_for_download in difference_all_vs_found.dropna(how="all")["all"]:
        left_to_download.append(file_for_download)

    return left_to_download


"""usage of what_is_left_to_download()"""
# left_to_download_mmCIF = what_is_left_to_download(input_mmCIF_files_were_found, all_mmCIF_file_names_files_from_latest_catalog)
# left_to_download_PDB = what_is_left_to_download(input_PDB_files_were_found, all_PDB_file_names_files_from_latest_catalog)
# left_to_download_SIFTS = what_is_left_to_download(input_SIFTS_files_were_found, all_SIFTS_file_names_files_from_latest_catalog)

# print(len(left_to_download_mmCIF))
# print(len(left_to_download_PDB))
# print(len(left_to_download_SIFTS))



# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import math
from collections import defaultdict
from Bio.PDB import calc_dihedral


from .table import extract_int, fix_val
from .lst import lst_nums
from .coord import resid_to_tuple, has_resid, has_atomid, get_atom_vect, calc_norm_dist, get_resname
from .col import (
    phi_col,
    psi_col,
    omega_col,
    chi1_col,
    chi2_col,
    chi3_col,
    chi4_col,
    chi5_col,
    altchi1_col,
    altchi2_col,
)

p_rot = 'p'
m_rot = 'm'
t_rot = 't'
n_rot = '-'

bb_dict = {
    phi_col: [["C", "N", "CA", "C"], [-1, 0, 0, 0]],
    psi_col: [["N", "CA", "C", "N"], [0, 0, 0, 1]],
    omega_col: [["CA", "C", "N", "CA"], [-1, -1, 0, 0]],
}

sc_dict = {
    chi1_col: {
        "ARG": ["N", "CA", "CB", "CG"],
        "ASN": ["N", "CA", "CB", "CG"],
        "ASP": ["N", "CA", "CB", "CG"],
        "CYS": ["N", "CA", "CB", "SG"],
        "GLN": ["N", "CA", "CB", "CG"],
        "GLU": ["N", "CA", "CB", "CG"],
        "HIS": ["N", "CA", "CB", "CG"],
        "ILE": ["N", "CA", "CB", "CG1"],
        "LEU": ["N", "CA", "CB", "CG"],
        "LYS": ["N", "CA", "CB", "CG"],
        "MET": ["N", "CA", "CB", "CG"],
        "PHE": ["N", "CA", "CB", "CG"],
        "PRO": ["N", "CA", "CB", "CG"],
        "SER": ["N", "CA", "CB", "OG"],
        "THR": ["N", "CA", "CB", "OG1"],
        "TRP": ["N", "CA", "CB", "CG"],
        "TYR": ["N", "CA", "CB", "CG"],
        "VAL": ["N", "CA", "CB", "CG1"],
    },
    altchi1_col: {"VAL": ["N", "CA", "CB", "CG2"]},
    chi2_col: {
        "ARG": ["CA", "CB", "CG", "CD"],
        "ASN": ["CA", "CB", "CG", "OD1"],
        "ASP": ["CA", "CB", "CG", "OD1"],
        "GLN": ["CA", "CB", "CG", "CD"],
        "GLU": ["CA", "CB", "CG", "CD"],
        "HIS": ["CA", "CB", "CG", "ND1"],
        "ILE": ["CA", "CB", "CG1", "CD1"],
        "LEU": ["CA", "CB", "CG", "CD1"],
        "LYS": ["CA", "CB", "CG", "CD"],
        "MET": ["CA", "CB", "CG", "SD"],
        "PHE": ["CA", "CB", "CG", "CD1"],
        "PRO": ["CA", "CB", "CG", "CD"],
        "TRP": ["CA", "CB", "CG", "CD1"],
        "TYR": ["CA", "CB", "CG", "CD1"],
    },
    altchi2_col: {
        "ASP": ["CA", "CB", "CG", "OD2"],
        "LEU": ["CA", "CB", "CG", "CD2"],
        "PHE": ["CA", "CB", "CG", "CD2"],
        "TYR": ["CA", "CB", "CG", "CD2"],
    },
    chi3_col: {
        "ARG": ["CB", "CG", "CD", "NE"],
        "GLN": ["CB", "CG", "CD", "OE1"],
        "GLU": ["CB", "CG", "CD", "OE1"],
        "LYS": ["CB", "CG", "CD", "CE"],
        "MET": ["CB", "CG", "SD", "CE"],
    },
    chi4_col: {
        "ARG": ["CG", "CD", "NE", "CZ"],
        "LYS": ["CG", "CD", "CE", "NZ"],
    },
    chi5_col: {"ARG": ["CD", "NE", "CZ", "NH1"]},
}


def get_rama_type(phi, psi, omega):

    phi = float(phi)
    psi = float(psi)
    omega = float(omega)

    omega = math.radians(omega)

    if phi > 180.0:
        phi = -360.0 + phi
    if psi > 180.0:
        psi = -360.0 + psi
    if phi == -180:
        phi = 180
    if psi == -180:
        psi = 180

    regions = defaultdict(list)

    regions["B"].append([-180, -100, 50, 180])
    regions["B"].append([-180, -100, -180, -100])
    regions["A"].append([-180, -100, -100, 50])
    regions["B"].append([-100, 0, -180, -100])
    regions["B"].append([-100, 0, 50, 180])
    regions["A"].append([-100, 0, -100, 50])
    regions["E"].append([0, 150, -180, -50])
    regions["E"].append([0, 150, 100, 180])
    regions["B"].append([150, 180, 100, 180])
    regions["B"].append([150, 180, -180, -50])
    regions["L"].append([0, 180, -50, 100])

    for region in regions:

        for rama in regions[region]:

            if (rama[0] < phi <= rama[1]) and (
                rama[2] < psi <= rama[3] and math.cos(omega) <= 0
            ):

                rama_type = region

            if (rama[0] < phi <= rama[1]) and (
                rama[2] < psi <= rama[3] and math.cos(omega) > 0
            ):

                region = region.lower()
                rama_type = region

    return rama_type


def get_rot_type(rot):

    rot = float(rot)

    if rot == 999.00:
        return n_rot
    elif 0 <= rot < 120:
        return p_rot
    elif 120 <= rot < 240:
        return t_rot
    elif 240 <= rot < 360:
        return m_rot 


def calc_dih_angle(
    structure,
    chainid,
    resid_1,
    resid_2,
    resid_3,
    resid_4,
    atomid_1,
    atomid_2,
    atomid_3,
    atomid_4,
    modelid=None,
    check_dist=True,
    check_resid=True,
    check_atomid=True,
    max_dist=2,
):

    if modelid is None:
        modelid = 0

    index_lst = lst_nums(0, 3)

    resid_lst = [resid_1, resid_2, resid_3, resid_4]
    atomid_lst = [atomid_1, atomid_2, atomid_3, atomid_4]

    for i, resid in enumerate(resid_lst):
        resid_lst[i] = resid_to_tuple(resid)

    structure_angle = 999.00

    calc_vect = True

    for index in index_lst:

        resid = resid_lst[index]
        atomid = atomid_lst[index]

        resid_status = True
        if check_resid:
            resid_status = has_resid(structure, chainid, resid, modelid=modelid)

        atomid_status = True
        if resid_status:
            if check_atomid:
                atomid_status = has_atomid(
                    structure, chainid, resid, atomid, modelid=modelid
                )

        if not resid_status or not atomid_status:
            calc_vect = False
            break

    if calc_vect:

        vect_lst = list()

        for index in index_lst:

            resid = resid_lst[index]
            atomid = atomid_lst[index]

            vect = get_atom_vect(
                structure=structure,
                modelid=modelid,
                chainid=chainid,
                resid=resid,
                atomid=atomid,
            )

            vect_lst.append(vect)

        calc_angle = True

        if check_dist:
            for index in index_lst:
                if index != 0 and index != 3:
                    atom_dist = calc_norm_dist(vect_lst[index], vect_lst[index + 1])
                    if atom_dist > max_dist:
                        calc_angle = False

        if calc_angle:
            structure_angle = round(
                math.degrees(
                    calc_dihedral(vect_lst[0], vect_lst[1], vect_lst[2], vect_lst[3])
                ),
                2,
            )

    return structure_angle


def calc_bb_angle(
    structure,
    chainid,
    curr_resid,
    angle,
    modelid=None,
    prev_resid=None,
    next_resid=None,
    check_dist=True,
    check_resid=True,
    check_atomid=True,
    max_dist=2,
):

    atomid_list = bb_dict[angle][0]
    add_list = bb_dict[angle][1]

    index_lst = lst_nums(0, 3)

    if type(curr_resid) == tuple:
        curr_resnum = curr_resid[1]
    else:
        curr_resnum = extract_int(curr_resid)
        curr_resid = resid_to_tuple(curr_resid)

    resid_lst = list()

    check_dist = False

    for index in index_lst:

        add = add_list[index]

        resid = curr_resid

        if add == -1:
            if prev_resid is not None:
                resid = prev_resid
                if type(prev_resid) == tuple:
                    prev_resnum = prev_resid[1]
                if curr_resnum - prev_resnum != 1:
                    check_dist = True
            else:
                resid = resid_to_tuple(curr_resid + add)
        elif add == 1:
            if next_resid is not None:
                resid = next_resid
                if type(next_resid) == tuple:
                    next_resnum = next_resid[1]
                if next_resnum - curr_resnum != 1:
                    check_dist = True
            else:
                resid = resid_to_tuple(curr_resid + add)

        resid_lst.append(resid)

    bb_angle = calc_dih_angle(
        structure=structure,
        modelid=modelid,
        chainid=chainid,
        resid_1=resid_lst[0],
        resid_2=resid_lst[1],
        resid_3=resid_lst[2],
        resid_4=resid_lst[3],
        atomid_1=atomid_list[0],
        atomid_2=atomid_list[1],
        atomid_3=atomid_list[2],
        atomid_4=atomid_list[3],
        check_dist=check_dist,
        check_resid=check_resid,
        check_atomid=check_atomid,
        max_dist=max_dist,
    )

    return bb_angle


def calc_sc_angle(
    structure,
    chainid,
    resid,
    angle,
    modelid=None,
    check_dist=True,
    check_resid=True,
    check_atomid=True,
    max_dist=2,
):

    if modelid is None:
        modelid = 0

    sc_angle = 999.00

    resid = resid_to_tuple(resid)

    if has_resid(structure, chainid, resid, modelid=modelid):

        residue = structure[fix_val(modelid, return_int=True)][chainid][resid]

        resname = get_resname(residue, letter=False)

        sc_angle_dict = sc_dict[angle]

        if resname in list(sc_angle_dict.keys()):

            atomid_lst = sc_dict[angle][resname]

            sc_angle = calc_dih_angle(
                structure=structure,
                modelid=modelid,
                chainid=chainid,
                resid_1=resid,
                resid_2=resid,
                resid_3=resid,
                resid_4=resid,
                atomid_1=atomid_lst[0],
                atomid_2=atomid_lst[1],
                atomid_3=atomid_lst[2],
                atomid_4=atomid_lst[3],
                check_dist=check_dist,
                check_resid=check_resid,
                check_atomid=check_atomid,
                max_dist=max_dist,
            )

            if sc_angle != 999.00:
                if (angle == chi2_col and resname in ["TYR", "PHE", "ASP"]) or (
                    angle == chi3_col and resname == "GLU"
                ):
                    if sc_angle < 0.00:
                        sc_angle += 180.00
                    elif sc_angle > 180.00:
                        sc_angle -= 180.00
                else:
                    if sc_angle < 0.00:
                        sc_angle += 360.00

    return sc_angle



from src.download.modules import *


def latest_catalog_reader():
    files_of_current_directory = os.listdir(current_directory)
    paths_to_ls_lR = list()
    paths_to_xml = list()

    for file_in_current_dir in files_of_current_directory:
        if file_in_current_dir.startswith('ls-lR'):
            file_in_current_dir = current_directory + "/" + file_in_current_dir + "/" + 'ls-lR'
            paths_to_ls_lR.append(file_in_current_dir)
        if file_in_current_dir.startswith('xml'):
            file_in_current_dir = current_directory + "/" + file_in_current_dir + "/" + 'xml'
            paths_to_xml.append(file_in_current_dir)

    paths_to_ls_lR_sorted = sorted(paths_to_ls_lR, reverse=True)
    paths_to_xml_sorted = sorted(paths_to_xml, reverse=True)

    try:
        path_to_the_latest_listing = paths_to_ls_lR_sorted[0]

        df_catalog_the_latest_listing = pd.read_csv(Path(path_to_the_latest_listing),
                                                    names=["1", "2", "3", "4", "Data_size", "Month", "Day", "Time",
                                                           "file_name", "10", "file_names_path"], sep="\s+",
                                                    low_memory=False)

        # mmCIF
        df_catalog_the_latest_mmCIF_listing_dropna = df_catalog_the_latest_listing.dropna()
        df_catalog_the_latest_mmCIF_listing_dropna_cif_gz = df_catalog_the_latest_mmCIF_listing_dropna[
            df_catalog_the_latest_mmCIF_listing_dropna['file_name'].str.endswith('cif.gz')]
        df_catalog_the_latest_mmCIF_listing_dropna_cif_gz_34kb = df_catalog_the_latest_mmCIF_listing_dropna_cif_gz[
            df_catalog_the_latest_mmCIF_listing_dropna_cif_gz.Data_size == 34.0]

        all_mmCIF_files = list()
        for mmCIF_file_name in df_catalog_the_latest_mmCIF_listing_dropna_cif_gz_34kb["file_name"]:
            all_mmCIF_files.append(mmCIF_file_name)

        # PDB
        df_catalog_the_latest_PDB_listing_dropna = df_catalog_the_latest_listing.dropna()
        df_catalog_the_latest_PDB_listing_dropna_ent_gz = df_catalog_the_latest_PDB_listing_dropna[
            df_catalog_the_latest_PDB_listing_dropna['file_name'].str.endswith('ent.gz')]
        df_catalog_the_latest_PDB_listing_dropna_ent_gz_34kb = df_catalog_the_latest_PDB_listing_dropna_ent_gz[
            df_catalog_the_latest_PDB_listing_dropna_ent_gz.Data_size == 35.0]

        all_PDB_files = list()
        for PDB_file_name in df_catalog_the_latest_PDB_listing_dropna_ent_gz_34kb["file_name"]:
            all_PDB_files.append(PDB_file_name)

        # SIFTS
        path_to_the_latest_list = paths_to_xml_sorted[0]
        df_catalog_the_latest_listing = pd.read_csv(path_to_the_latest_list, header=None, sep="\s+", low_memory=False)
        all_SIFTS_files = set()
        for key in df_catalog_the_latest_listing:
            try:
                if df_catalog_the_latest_listing[key].str.contains(".xml.gz").any():
                    for val in df_catalog_the_latest_listing[key]:
                        if val.endswith(".xml.gz"):
                            if "/" in val:
                                all_SIFTS_files.add(val.split("/")[-1])
                            else:
                                all_SIFTS_files.add(val)
            except AttributeError:
                pass

    except IndexError:
        print("Sorry, nothing to read from. Try catalog_downloader() command first.")
        all_mmCIF_files = None
        all_PDB_files = None
        all_SIFTS_files = None

    return [all_mmCIF_files, all_PDB_files, all_SIFTS_files]


"""usage of latest_catalog_reader()"""
# all_data_list_reader = latest_catalog_reader()
# all_mmCIF_file_names_files_from_latest_catalog = all_data_list_reader[0]
# all_PDB_file_names_files_from_latest_catalog = all_data_list_reader[1]
# all_SIFTS_file_names_files_from_latest_catalog = all_data_list_reader[2]

# print(len(all_mmCIF_file_names_files_from_latest_catalog))
# print(len(all_PDB_file_names_files_from_latest_catalog))
# print(len(all_SIFTS_file_names_files_from_latest_catalog))



# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import pandas as pd
from tqdm import tqdm

from ..functions.lst import str_to_lst, lst_to_str, sort_lst, res_to_lst
from ..functions.table import extract_int, fix_val, get_df_at_index, fix_col
from ..functions.coord import (
    load_coord,
    get_neighbors,
    has_resid,
    resid_to_tuple,
    get_resname,
    get_residue_cont,
    get_reschainid,
    get_resnum,
    is_aa,
    calc_atom_dist,
)
from ..functions.col import (
    interf_path_col,
    atomid_cont_col,
    renum_path_col,
    modelid_col,
    chainid_col,
    interf_area_col,
    iso_col,
    interf_col,
    interf_cont_col,
    cb_dist_col,
    interf_id_col,
    pdb_id_col,
)
from ..functions.interf import calc_q_score
from ..functions.path import save_table


def calc_interf_cont_dist(
    coord_path,
    chainid,
    interf,
    interf_dict,
    cont_dist=5,
    modelid=None,
    interf_resid_lst=None,
):

    if modelid is None:
        modelid = 0

    interf_path = interf_dict[coord_path][chainid][interf][interf_path_col]

    structure = load_coord(interf_path)

    neighbors = get_neighbors(structure)

    resid_lst = str_to_lst(interf_dict[coord_path][chainid][interf][atomid_cont_col])

    interf_cont_lst = list()
    cb_dist_lst = list()

    add_interf = True
    if interf_resid_lst is not None:
        add_interf = False

    for resid in resid_lst:

        if interf_resid_lst is not None:
            if extract_int(resid) in interf_resid_lst:
                add_interf = True

        if resid is not None and add_interf:

            if has_resid(structure, chainid, resid_to_tuple(resid), modelid=modelid):
                residue = structure[fix_val(modelid, return_int=True)][chainid][
                    resid_to_tuple(resid)
                ]

                resname = get_resname(residue)

                if resname == "GLY":
                    atomid = "CA"
                else:
                    atomid = "CB"

                for cont_residue in get_residue_cont(
                    neighbors, residue, max_dist=cont_dist, level="R"
                ):
                    cont_chainid = get_reschainid(cont_residue)
                    cont_resid = get_resnum(cont_residue)
                    cont_resname = get_resname(cont_residue)

                    if is_aa(cont_residue) and cont_chainid != chainid:

                        if cont_resname == "GLY":
                            cont_atomid = "CA"
                        else:
                            cont_atomid = "CB"

                        if cont_resid is not None:

                            interf_cont = lst_to_str(
                                sort_lst([resid, cont_resid]),
                                join_txt=":",
                            )

                            if interf_cont not in interf_cont_lst:

                                cb_dist = calc_atom_dist(
                                    structure,
                                    modelid_1=modelid,
                                    chainid_1=chainid,
                                    resid_1=resid,
                                    atomid_1=atomid,
                                    modelid_2=modelid,
                                    chainid_2=cont_chainid,
                                    resid_2=cont_resid,
                                    atomid_2=cont_atomid,
                                )

                                interf_cont_lst.append(interf_cont)
                                cb_dist_lst.append(cb_dist)

    return interf_cont_lst, cb_dist_lst


def get_index_interf(
    df,
    index,
    interf_dict,
    interf_resid_lst=None,
    min_area=200,
    cont_dist=5,
    iso_interf=False,
    het_interf=False,
    search_interf_cont_lst=None,
    search_cb_dist_lst=None,
    search_max_dist=0.7,
    coord_path_col=None,
):

    if coord_path_col is None:
        coord_path_col = renum_path_col

    index_df = get_df_at_index(df, index)

    coord_path = index_df.at[index, coord_path_col]
    modelid = index_df.at[index, modelid_col]
    chainid = index_df.at[index, chainid_col]

    interf_df = pd.DataFrame()

    for interf in list(interf_dict[coord_path][chainid].keys()):

        add_interf = True

        if float(interf_dict[coord_path][chainid][interf][interf_area_col]) < min_area:
            add_interf = False

        interf_iso_status = interf_dict[coord_path][chainid][interf][iso_col]
        if iso_interf:
            if not interf_iso_status:
                add_interf = False
        if het_interf:
            if interf_iso_status:
                add_interf = False

        if add_interf:

            interf_cont_lst, cb_dist_lst = calc_interf_cont_dist(
                coord_path,
                chainid,
                interf,
                interf_dict,
                cont_dist=cont_dist,
                modelid=modelid,
                interf_resid_lst=interf_resid_lst,
            )

            if search_interf_cont_lst is not None and search_cb_dist_lst is not None:
                if (
                    calc_q_score(
                        i_cont_lst=search_interf_cont_lst,
                        j_cont_lst=interf_cont_lst,
                        i_dist_lst=search_cb_dist_lst,
                        j_dist_lst=cb_dist_lst,
                    )
                    > search_max_dist
                ):
                    add_interf = False

        if add_interf:
            temp_df = index_df.copy(deep=True)

            for col in list(interf_dict[coord_path][chainid][interf].keys()):
                temp_df.at[index, col] = interf_dict[coord_path][chainid][interf][col]

            temp_df.at[index, interf_col] = interf
            temp_df.at[index, interf_cont_col] = lst_to_str(interf_cont_lst)
            temp_df.at[index, cb_dist_col] = lst_to_str(cb_dist_lst)

            interf_df = pd.concat([interf_df, temp_df], sort=False)

    return interf_df


def build_interf_table(
    df,
    interf_dict,
    interf_table_path=None,
    interf_resids=None,
    min_area=200,
    cont_dist=5,
    iso_interf=False,
    het_interf=False,
    search_coord_path=None,
    search_modelid=None,
    search_chainid=None,
    search_interf=None,
    search_max_dist=0.7,
    coord_path_col=None,
):

    df = df.reset_index(drop=True)

    if coord_path_col is None:
        coord_path_col = renum_path_col

    interf_df = pd.DataFrame()

    interf_resid_lst = res_to_lst(interf_resids)

    search_interf_cont_lst = None
    search_cb_dist_lst = None

    if (
        search_coord_path is not None
        and search_chainid is not None
        and search_interf is not None
    ):

        try:
            search_interf_cont_lst, search_cb_dist_lst = calc_interf_cont_dist(
                search_coord_path,
                search_chainid,
                str(search_interf),
                interf_dict,
                cont_dist=cont_dist,
                modelid=search_modelid,
                interf_resid_lst=interf_resid_lst,
            )
        except:
            search_interf_cont_lst, search_cb_dist_lst = calc_interf_cont_dist(
                search_coord_path,
                search_chainid,
                int(search_interf),
                interf_dict,
                cont_dist=cont_dist,
                modelid=search_modelid,
                interf_resid_lst=interf_resid_lst,
            )

    for index in tqdm(
        list(df.index.values), desc="Building interface table", position=0, leave=True
    ):

        interf_df = pd.concat(
            [
                interf_df,
                get_index_interf(
                    df,
                    index,
                    interf_dict,
                    min_area=min_area,
                    cont_dist=cont_dist,
                    interf_resid_lst=interf_resid_lst,
                    iso_interf=iso_interf,
                    het_interf=het_interf,
                    search_interf_cont_lst=search_interf_cont_lst,
                    search_cb_dist_lst=search_cb_dist_lst,
                    search_max_dist=search_max_dist,
                    coord_path_col=coord_path_col,
                ),
            ],
            sort=False,
        )

    if len(interf_df) > 0:
        interf_df = interf_df.reset_index(drop=True)

        interf_df[interf_col] = interf_df[interf_col].map(str)
        interf_df = fix_col(interf_df, interf_col)

        interf_df[interf_id_col] = interf_df[pdb_id_col] + interf_df[interf_col]

    if interf_table_path is not None:
        save_table(interf_table_path, interf_df)
    else:
        return interf_df

    print("Built interface table!")



# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import time
import pandas as pd
import requests
import json
from tqdm import tqdm
import concurrent.futures

from ..functions.table import merge_dicts
from ..functions.lst import type_lst
from ..functions.download import download_file
from ..functions.path import (
    path_exists,
    save_table,
    get_edia_path,
    append_path,
    get_dir_path,
    save_json,
    edia_str,
)
from ..functions.col import edia_col, b_factor_col


def download_edia_scores(path, pdb_code):

    if not path_exists(path):

        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
        }

        data = '{"edia": {"pdbCode": "%s"}}' % pdb_code

        response = requests.post(
            "https://proteins.plus/api/edia_rest", headers=headers, data=data
        )

        wait = 1
        while True:
            try:
                data = json.loads(response.text)
                status = data["status_code"]
                break
            except Exception:
                time.sleep(wait)
                wait *= 1.5

        if status != 400:

            url = data["location"]

            wait = 1
            while True:
                response = requests.get(url)
                data = json.loads(response.text)
                status = data["status_code"]

                if status == 200:
                    url = data["atom_scores"]
                    download_file(url, path)
                    break
                elif status == 400:
                    message = data["message"]
                    if message == "No electron density file available.":
                        df = pd.DataFrame(
                            columns=[
                                "Structure specifier",
                                "Atom name",
                                "Infile id",
                                "Substructure name",
                                "Substructure id",
                                "Chain",
                                "Element",
                                "EDIA",
                                "EDIA fault analysis",
                                "B factor",
                                "Occupancy",
                            ]
                        )
                        save_table(path, df, sep=",", header=True, index=False)
                        break
                elif status == 202 or status == 429:
                    time.sleep(wait)
                    wait *= 1.5


def get_pdb_edia_dict(pdb_code, edia_dir=None, sifts_dict=None):

    edia_path = get_edia_path(pdb_code, dir_path=edia_dir)

    download_edia_scores(edia_path, pdb_code)

    df = pd.read_csv(edia_path, header=0)

    edia_dict = {pdb_code: dict()}

    if len(df) != 0:

        for index in list(df.index.values):

            add_val = True

            chainid = df.at[index, "Chain"]
            resnum = str(df.at[index, "Substructure id"])
            atomid = df.at[index, "Atom name"]
            edia = df.at[index, "EDIA"]
            b_factor = df.at[index, "B factor"]

            if chainid not in list(edia_dict[pdb_code].keys()):
                edia_dict[pdb_code][chainid] = dict()

            if sifts_dict is not None:
                if resnum in list(sifts_dict[pdb_code][chainid].keys()):
                    resnum = sifts_dict[pdb_code][chainid][str(resnum)]

                else:
                    add_val = False

            if add_val:

                if resnum not in list(edia_dict[pdb_code][chainid].keys()):

                    edia_dict[pdb_code][chainid][resnum] = dict()

                edia_dict[pdb_code][chainid][resnum][atomid] = {
                    edia_col: edia,
                    b_factor_col: b_factor,
                }

    return edia_dict


def prep_edia(
    pdb_codes,
    edia_dir=None,
    edia_json_path=None,
    sifts_dict=None,
    num_cpu=1,
):

    append_path(get_dir_path(dir_str=edia_str, dir_path=edia_dir))

    pdb_code_lst = type_lst(pdb_codes)

    edia_dict = dict()

    if num_cpu == 1:
        for pdb_code in tqdm(
            pdb_code_lst,
            desc="Preparing EDIA scores",
            position=0,
            leave=True,
        ):

            edia_dict = merge_dicts(
                [
                    edia_dict,
                    get_pdb_edia_dict(
                        pdb_code, edia_dir=edia_dir, sifts_dict=sifts_dict
                    ),
                ]
            )
    else:
        with concurrent.futures.ProcessPoolExecutor(max_workers=num_cpu) as executor:
            job_lst = [
                executor.submit(
                    get_pdb_edia_dict,
                    pdb_code,
                    edia_dir=edia_dir,
                    sifts_dict=sifts_dict,
                )
                for pdb_code in pdb_code_lst
            ]

            for job in tqdm(
                concurrent.futures.as_completed(job_lst),
                desc="Prepared EDIA scores",
                total=len(job_lst),
                miniters=1,
                position=0,
                leave=True,
            ):

                edia_dict = merge_dicts([edia_dict, job.result()])

    print("Prepared EDIA scores!")

    if edia_json_path is not None:
        save_json(edia_json_path, edia_dict)
    else:
        return edia_dict



# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import numpy as np
from tqdm import tqdm
import itertools

from ..functions.table import order_rows
from ..functions.lst import str_to_lst, calc_simpson, calc_jaccard
from ..functions.col import pocket_cont_col
from ..functions.path import save_matrix


def build_pocket_matrix(fit_df, pocket_matrix_path, pred_df=None, use_simpson=False):

    fit_df = order_rows(fit_df)

    fit_df = order_rows(fit_df)
    j_df = fit_df.copy(deep=True)

    if pred_df is None:
        i_df = fit_df.copy(deep=True)
    else:
        pred_df = order_rows(pred_df)
        i_df = pred_df.copy(deep=True)

    i_index_lst = list(i_df.index.values)
    j_index_lst = list(j_df.index.values)

    if pred_df is not None:
        index_pairs = itertools.product(i_index_lst, j_index_lst)
    else:
        index_pairs = itertools.combinations(i_index_lst, 2)

    matrix = np.zeros((len(i_index_lst), len(j_index_lst)))

    for i_index, j_index in tqdm(
        index_pairs, desc="Building pocket matrix", position=0, leave=True
    ):

        i_cont_lst = str_to_lst(i_df.at[i_index, pocket_cont_col])
        j_cont_lst = str_to_lst(j_df.at[j_index, pocket_cont_col])

        if use_simpson:
            dist = calc_simpson(
                i_cont_lst,
                j_cont_lst,
                return_dist=True,
            )

        else:
            dist = calc_jaccard(
                i_cont_lst,
                j_cont_lst,
                return_dist=True,
            )

        matrix[i_index, j_index] = dist

        if pred_df is None:
            matrix[j_index, i_index] = dist

    save_matrix(pocket_matrix_path, matrix)

    print("Built pocket matrix!")